from functools import partial
from itertools import combinations
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from scipy.special import comb
from tabulate import tabulate

from bratiaa.evaluation import *
from bratiaa.utils import read, TokenOverlap
from bratsubset.projectconfig import ProjectConfiguration

# attempted division by zero is expected and unproblematic -> NaN
np.seterr(divide='ignore', invalid='ignore')

AnnFile = namedtuple('AnnFile', ['annotator_id', 'ann_path'])


class Document:
    __slots__ = ['ann_files', 'txt_path', 'doc_id']

    def __init__(self, txt_path, doc_id=None):
        self.txt_path = txt_path
        if doc_id:
            self.doc_id = doc_id
        else:
            self.doc_id = txt_path
        self.ann_files = []


def input_generator(root):
    """
    Yields Document objects. Assumes that each first-level subdirectory of the
    annotation project corresponds to one annotator.
    """
    root = Path(root)
    annotators = [subdir.parts[-1] for subdir in root.glob('*/') if subdir.is_dir()]
    assert len(annotators) > 1, 'At least two annotators are necessary to compute agreement!'
    for rel_path in sorted(collect_redundant_files(root, annotators)):
        document = Document((root / annotators[0] / rel_path).as_posix()[:-3] + 'txt', doc_id=rel_path)
        for annotator in annotators:
            ann_path = root / annotator / rel_path
            document.ann_files.append(AnnFile(annotator, ann_path))
        yield document


def collect_redundant_files(root, annotators):
    intersection = None
    for annotator in annotators:
        subdir_path = root / annotator
        relative_paths = {path.relative_to(subdir_path).as_posix() for path in subdir_path.glob('**/*.ann')}
        if not intersection:
            intersection = relative_paths
        else:
            intersection.intersection(relative_paths)
    return intersection


def _collect_annotators_and_documents(input_gen):
    annotators, documents = set(), []
    for document in input_gen():
        for ann_file in document.ann_files:
            annotators.add(ann_file.annotator_id)
        documents.append(document.doc_id)
    return list(annotators), documents


def compute_f1(tp, fp, fn):
    return (2 * tp) / (2 * tp + fp + fn)


class F1Agreement:
    def __init__(self, input_gen, labels, eval_func=exact_match_instance_evaluation, token_func=None, annotators=None,
                 documents=None):
        if not (annotators and documents):
            annotators, documents = _collect_annotators_and_documents(input_gen)
            annotators.sort()
            documents.sort()
        assert len(annotators) > 1, 'At least two annotators are necessary to compute agreement!'
        num_pairs = comb(len(annotators), 2, exact=True)
        # (p, d, c, l) where p := annotator pairs, d := documents, c := counts (tp, fp, fn), l := labels
        self._pdcl = np.zeros((num_pairs, len(documents), 3, len(labels)))
        self._documents = list(documents)
        self._doc2idx = {d: i for i, d in enumerate(documents)}
        self._labels = list(labels)
        self._label2idx = {l: i for i, l in enumerate(labels)}
        self._annotators = list(annotators)
        self._pairs = [pair for pair in combinations(annotators, 2)]
        self._pair2idx = {p: i for i, p in enumerate(self._pairs)}
        # add pairs in reverse order (same index)
        for (a1, a2), value in self._pair2idx.copy().items():
            self._pair2idx[(a2, a1)] = value
        self._eval_func = eval_func  # function used to extract true positives, false positives and false negatives
        self._token_func = token_func  # function used for tokenization
        self._compute_tp_fp_fn(input_gen)

    @property
    def annotators(self):
        return list(self._annotators)

    @property
    def documents(self):
        return list(self._documents)

    @property
    def labels(self):
        return list(self._labels)

    def _compute_tp_fp_fn(self, input_gen):
        for doc_index, document in enumerate(input_gen()):
            assert doc_index < len(self._documents), 'Input generator yields more documents than expected!'
            to = None
            if self._token_func:
                text = read(document.txt_path)
                tokens = list(self._token_func(text))
                to = TokenOverlap(text, tokens)
            for anno_file_1, anno_file_2 in combinations(document.ann_files, 2):
                tp, fp, fn = self._eval_func(anno_file_1.ann_path, anno_file_2.ann_path, tokens=to)
                pair_idx = self._pair2idx[(anno_file_1.annotator_id, anno_file_2.annotator_id)]
                doc_idx = self._doc2idx[document.doc_id]
                self._increment_counts(tp, pair_idx, doc_idx, 0)
                self._increment_counts(fp, pair_idx, doc_idx, 1)
                self._increment_counts(fn, pair_idx, doc_idx, 2)

    def _increment_counts(self, annotations, pair, doc, kind):
        for a in annotations:
            self._pdcl[pair][doc][kind][self._label2idx[a.label]] += 1

    def mean_sd_per_label(self):
        """
        Mean and standard deviation of all annotator combinations' F1 scores by label.
        """
        pcl = np.sum(self._pdcl, axis=1)  # sum over documents
        f1_pairs = compute_f1(pcl[:, 0], pcl[:, 1], pcl[:, 2])
        avg, stddev = self._mean_sd(f1_pairs)
        return avg, stddev

    def mean_sd_per_document(self):
        """
        Mean and standard deviation of all annotator combinations' F1 scores per document.
        """
        pdc = np.sum(self._pdcl, axis=3)  # sum over labels
        f1_pairs = compute_f1(pdc[:, :, 0], pdc[:, :, 1], pdc[:, :, 2])
        avg, stddev = self._mean_sd(f1_pairs)
        return avg, stddev

    def mean_sd_total(self):
        """
        Mean and standard deviation of all annotator cominations' F1 scores.
        """
        pc = np.sum(self._pdcl, axis=(1, 3))  # sum over documents and labels
        f1_pairs = compute_f1(pc[:, 0], pc[:, 1], pc[:, 2])
        avg, stddev = self._mean_sd(f1_pairs)
        return avg, stddev

    def mean_sd_per_label_one_vs_rest(self, annotator):
        """
        Mean and standard deviation of all annotator combinations' F1 scores involving given annotator per label.
        """
        pcl = np.sum(self._pdcl, axis=1)  # sum over documents
        pcl = pcl[self._pairs_involving(annotator)]
        f1_pairs = compute_f1(pcl[:, 0], pcl[:, 1], pcl[:, 2])
        avg, stddev = self._mean_sd(f1_pairs)
        return avg, stddev

    def mean_sd_total_one_vs_rest(self, annotator):
        """
        Mean and standard deviation of all annotator combinations' F1 scores involving given annotator.
        """
        pc = np.sum(self._pdcl, axis=(1, 3))  # sum over documents and labels
        pc = pc[self._pairs_involving(annotator)]
        f1_pairs = compute_f1(pc[:, 0], pc[:, 1], pc[:, 2])
        if len(f1_pairs) > 1:
            avg, stddev = self._mean_sd(f1_pairs)
        else:
            avg, stddev = f1_pairs, 0
        return avg, stddev

    def _pairs_involving(self, annotator):
        return [self._pair2idx[(a1, a2)] for (a1, a2) in self._pairs if
                a1 == annotator or a2 == annotator]

    @staticmethod
    def _mean_sd(f1_pairs):
        """
        Mean and standard deviation along first axis.
        """
        f1 = np.mean(f1_pairs, axis=0)
        f1_std = np.std(f1_pairs, axis=0)
        return f1, f1_std

    @staticmethod
    def print_table(row_label_header, row_labels, avg, stddev, precision=3):
        stats = np.stack((row_labels, avg, stddev)).transpose()
        headers = [row_label_header, 'Mean F1', 'SD F1']
        print(tabulate(stats, headers=headers, tablefmt='github', floatfmt=f'.{precision}f'))

    def compute_total_f1_matrix(self):
        """
        Returns (n x n) matrix, where n is the number of annotators, containing
        pair-wise total F1 scores between all annotators.

        By definition, the matrix is symmetric and F1 = 1 on the main diagonal.
        """
        pc = np.sum(self._pdcl, axis=(1, 3))  # sum over documents and labels
        f1_pairs = compute_f1(pc[:, 0], pc[:, 1], pc[:, 2])
        num_annotators = len(self._annotators)
        f1_matrix = np.zeros((num_annotators, num_annotators))
        annotator2idx = {a: i for i, a in enumerate(self._annotators)}
        for ann1, ann2 in self._pairs:
            ann1_idx, ann2_idx = annotator2idx[ann1], annotator2idx[ann2]
            f1_matrix[ann1_idx][ann2_idx] = f1_pairs[self._pair2idx[(ann1, ann2)]]
            f1_matrix[ann2_idx][ann1_idx] = f1_pairs[self._pair2idx[(ann2, ann1)]]
        # perfect diagonal by definition
        for i in range(len(self._annotators)):
            f1_matrix[i][i] = 1
        return f1_matrix

    def draw_heatmap(self, out_path):
        """
        Draws heatmap based on square matrix of F1 scores.
        """
        matrix = self.compute_total_f1_matrix()
        fig, ax = plt.subplots()
        im = ax.imshow(matrix)
        # ticks and labels
        num_annotators = len(self._annotators)
        ax.set_xticks(np.arange(num_annotators))
        ax.set_yticks(np.arange(num_annotators))
        # show complete grid
        ax.set_ylim(-0.5, num_annotators - 0.5)
        ax.set_ylim(num_annotators - 0.5, -0.5)
        ax.set_xticklabels(self._annotators)
        ax.set_yticklabels(self._annotators)
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                 rotation_mode="anchor")
        for i in range(len(self._annotators)):
            for j in range(len(self._annotators)):
                ax.text(j, i, f'{matrix[i, j]:.2f}', ha="center", va="center", color="w")

        # color bar
        cbar = ax.figure.colorbar(im, ax=ax)
        cbar.ax.set_ylabel('F1 score', rotation=-90, va="bottom")

        fig.tight_layout()
        plt.savefig(out_path)


def compute_f1_agreement(project_root, input_gen=input_generator, token_func=None, eval_func=None):
    if not eval_func:
        eval_func = exact_match_instance_evaluation
        if token_func:
            eval_func = exact_match_token_evaluation

    config = ProjectConfiguration(project_root)
    labels = config.get_entity_types()
    input_gen = partial(input_gen, project_root)
    annotators, documents = _collect_annotators_and_documents(input_gen)

    return F1Agreement(input_gen, sorted(labels), eval_func=eval_func, token_func=token_func,
                       annotators=sorted(annotators),
                       documents=sorted(documents))


def iaa_report(f1_agreement, precision=3):
    agreement_type = '* Instance-based F1 agreement'
    if f1_agreement._token_func:
        agreement_type = '* Token-based F1 agreement'

    print(f'# Inter-Annotator Agreement Report\n')

    print(agreement_type)

    print('\n## Project Setup\n')
    print(f'* {len(f1_agreement.annotators)} annotators: {", ".join(f1_agreement.annotators)}')
    print(f'* {len(f1_agreement.documents)} agreement documents')
    print(f'* {len(f1_agreement.labels)} labels')

    print('\n## Agreement per Document\n')
    f1_agreement.print_table('Document', f1_agreement.documents, *f1_agreement.mean_sd_per_document(),
                             precision=precision)

    print('\n## Agreement per Label\n')
    f1_agreement.print_table('Label', f1_agreement.labels, *f1_agreement.mean_sd_per_label(), precision=precision)

    print('\n## Overall Agreement\n')
    avg, stddev = f1_agreement.mean_sd_total()
    print(f'* Mean F1: {avg:.{precision}f}, SD F1: {stddev:.{precision}f}\n')



"""
Functions for computing the difference between two sets of annotations.
"""
from collections import namedtuple, Counter

import bratsubset.annotation as bs

Annotation = namedtuple('Annotation', ['type', 'label', 'offsets'])


def exact_match_instance_evaluation(ann_path_1, ann_path_2, tokens=None):
    exp = set(_read_textbound_annotations(ann_path_1))
    pred = set(_read_textbound_annotations(ann_path_2))
    tp = exp.intersection(pred)
    fp = pred.difference(exp)
    fn = exp.difference(pred)
    return tp, fp, fn


def _read_textbound_annotations(ann_path):
    with bs.Annotations(ann_path.as_posix(), read_only=True) as annotations:
        for annotation in annotations.get_textbounds():
            yield Annotation('T', annotation.type, tuple(annotation.spans))


def exact_match_token_evaluation(ann_path_1, ann_path_2, tokens=None):
    """
    Annotations are split into token-sized bits before true positives, false positives and false negatives are computed.

    Sub-token annotations are expanded to full tokens. Long annotations will influence the results more than short
    annotations. Boundary errors for adjacent annotations with the same label are ignored!
    """
    exp = Counter(_read_token_annotations(ann_path_1, tokens))
    pred = Counter(_read_token_annotations(ann_path_2, tokens))
    tp = counter2list(exp & pred)
    fp = counter2list(pred - exp)
    fn = counter2list(exp - pred)
    return tp, fp, fn


def counter2list(c):
    for elem, cnt in c.items():
        for i in range(cnt):
            yield (elem)


def _read_token_annotations(ann_path, tokens):
    """
    Yields a new annotation for each token overlapping with an annotation. If annotations are overlapping each other,
    there will be multiple annotations for a single token.
    """
    for annotation in set(_read_textbound_annotations(ann_path)):
        for start, end in annotation.offsets:
            for ts, te in tokens.overlapping_tokens(start, end):
                yield Annotation(annotation.type, annotation.label, ((ts, te),))



#!/usr/bin/env python
# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; coding: utf-8; -*-
# vim:set ft=python ts=4 sw=4 sts=4 autoindent:

"""Functionality shared between server components.

Author:     Pontus Stenetorp    <pontus is s u-tokyo ac jp>
Version:    2011-04-21
"""


class ProtocolError(Exception):
    def __init__(self):
        pass

    def __str__(self):
        # TODO: just adding __str__ to ProtocolError, not all
        # currently support it, so falling back on this assumption
        # about how to make a (nearly) human-readable string. Once
        # __str__ added to all ProtocolErrors, raise
        # NotImplementedError instead.
        return 'ProtocolError: %s (TODO: __str__() method)' % self.__class__

    def json(self, json_dic):
        raise NotImplementedError('abstract method')


class ProtocolArgumentError(ProtocolError):
    def json(self, json_dic):
        json_dic['exception'] = 'protocolArgumentError'

# If received by ajax.cgi, no JSON will be sent
# XXX: This is an ugly hack to circumvent protocol flaws


class NoPrintJSONError(Exception):
    def __init__(self, hdrs, data):
        self.hdrs = hdrs
        self.data = data


class NotImplementedError(ProtocolError):
    def json(self, json_dic):
        json_dic['exception'] = 'notImplemented'


class CollectionNotAccessibleError(ProtocolError):
    def json(self, json_dic):
        json_dic['exception'] = 'collectionNotAccessible'

    def __str__(self):
        return 'Error: collection not accessible'

# TODO: We have issues using this in relation to our inspection
#       in dispatch, can we make it work?
# Wrapper to send a deprecation warning to the client if debug is set


def deprecated_action(func):
    try:
        from config import DEBUG
    except ImportError:
        DEBUG = False
    from functools import wraps
    from message import Messager

    @wraps(func)
    def wrapper(*args, **kwds):
        if DEBUG:
            Messager.warning(
                ('Client sent "%s" action '
                 'which is marked as deprecated') %
                func.__name__,)
        return func(*args, **kwds)
    return wrapper

# relpath is not included in python 2.5; alternative implementation from
# BareNecessities package, License: MIT, Author: James Gardner
# TODO: remove need for relpath instead


def relpath(path, start):
    """Return a relative version of a path."""
    from os.path import abspath, sep, pardir, commonprefix
    from os.path import join as path_join
    if not path:
        raise ValueError("no path specified")
    start_list = abspath(start).split(sep)
    path_list = abspath(path).split(sep)
    # Work out how much of the filepath is shared by start and path.
    i = len(commonprefix([start_list, path_list]))
    rel_list = [pardir] * (len(start_list) - i) + path_list[i:]
    if not rel_list:
        return path
    return path_join(*rel_list)



#!/usr/bin/env python
# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; coding: utf-8; -*-
# vim:set ft=python ts=4 sw=4 sts=4 autoindent:

"""Wrapper for safely importing Messager with a fallback that will get
_something_ to the user even if Messager itself breaks."""

try:
    from bratsubset.realmessage import Messager
except BaseException:
    from bratsubset.sosmessage import SosMessager as Messager



#!/usr/bin/env python
# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; coding: utf-8; -*-
# vim:set ft=python ts=4 sw=4 sts=4 autoindent:

"""Server-to-client messaging-related functionality for Brat Rapid Annotation
Tool (brat)

NOTE: This module is used by ajax.cgi prior to verifying that the Python
version is new enough to run with all our other modules. Thus this module has
to be kept as backwards compatible as possible and this over-rides any
requirements on style otherwise imposed on the project.

Author:     Pontus Stenetorp    <pontus is s u-tokyo ac jp>
Author:     Sampo Pyysalo       <smp is s u-tokyo ac jp>
Version:    2011-05-31
"""

import re

# for cleaning up control chars from a string, from
# http://stackoverflow.com/questions/92438/stripping-non-printable-characters-from-a-string-in-python
# allow tab (9) and [unix] newline (10)
__control_chars = ''.join(
    map(chr, list(range(0, 9)) + list(range(11, 32)) + list(range(127, 160))))
__control_char_re = re.compile('[%s]' % re.escape(__control_chars))


def remove_control_chars(s):
    return __control_char_re.sub('', s)


class Messager:
    __pending_messages = []

    def info(msg, duration=3, escaped=False):
        Messager.__message(msg, 'comment', duration, escaped)
    # decorator syntax only since python 2.4, staticmethod() since 2.2
    info = staticmethod(info)

    def warning(msg, duration=3, escaped=False):
        Messager.__message(msg, 'warning', duration, escaped)
    warning = staticmethod(warning)

    def error(msg, duration=3, escaped=False):
        Messager.__message(msg, 'error', duration, escaped)
    error = staticmethod(error)

    def debug(msg, duration=3, escaped=False):
        Messager.__message(msg, 'debug', duration, escaped)
    debug = staticmethod(debug)

    def output(o):
        for m, c, d in Messager.__pending_messages:
            print(c, ":", m, file=o)
    output = staticmethod(output)

    def output_json(json_dict):
        try:
            return Messager.__output_json(json_dict)
        except Exception as e:
            # TODO: do we want to always give the exception?
            json_dict['messages'] = [
                [
                    'Messager error adding messages to json (internal error in message.py, please contact administrator): %s' %
                    str(e), 'error', -1]]
            return json_dict
    output_json = staticmethod(output_json)

    def __output_json(json_dict):
        # protect against non-unicode inputs
        convertable_messages = []
        for m in Messager.__pending_messages:
            try:
                m[0].encode('utf-8')
                convertable_messages.append(m)
            except UnicodeDecodeError:
                convertable_messages.append(
                    ('[ERROR: MESSAGE THAT CANNOT BE ENCODED AS UTF-8 OMITTED]', 'error', 5))
        Messager.__pending_messages = convertable_messages

        # clean up messages by removing possible control characters
        # that may cause trouble clientside
        cleaned_messages = []
        for s, t, r in Messager.__pending_messages:
            cs = remove_control_chars(s)
            if cs != s:
                s = cs + \
                    '[NOTE: SOME NONPRINTABLE CHARACTERS REMOVED FROM MESSAGE]'
            cleaned_messages.append((s, t, r))
        Messager.__pending_messages = cleaned_messages

        # to avoid crowding the interface, combine messages with identical
        # content
        msgcount = {}
        for m in Messager.__pending_messages:
            msgcount[m] = msgcount.get(m, 0) + 1

        merged_messages = []
        for m in Messager.__pending_messages:
            if m in msgcount:
                count = msgcount[m]
                del msgcount[m]
                s, t, r = m
                if count > 1:
                    s = s + '<br/><b>[message repeated %d times]</b>' % count
                merged_messages.append((s, t, r))

        if 'messages' not in json_dict:
            json_dict['messages'] = []
        json_dict['messages'] += merged_messages
        Messager.__pending_messages = []
        return json_dict
    __output_json = staticmethod(__output_json)

    def __escape(msg):
        from html import escape
        return escape(msg).replace('\n', '\n<br/>\n')
    __escape = staticmethod(__escape)

    def __message(msg, type, duration, escaped):
        if not isinstance(msg, str) and not isinstance(msg, str):
            msg = str(msg)
        if not escaped:
            msg = Messager.__escape(msg)
        Messager.__pending_messages.append((msg, type, duration))
    __message = staticmethod(__message)


if __name__ == '__main__':
    # Try out Unicode, that is always fun
    Messager.warning('Hello 世界！')
    json_dic = {}
    Messager.output_json(json_dic)
    print(json_dic)



import logging

import argparse

from bratiaa.agree import iaa_report, compute_f1_agreement
from bratiaa.utils import tokenize


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('project_root',
                        help='Root directory of the Brat annotation project')
    parser.add_argument('--heatmap',
                        help='Output path for F1-agreement heatmap',
                        dest='heatmap_path')
    parser.add_argument('-p', '--precision',
                        help='Precision of results (number of digits following the decimal point)',
                        dest='precision',
                        default=3)
    parser.add_argument('-s', '--silent',
                        help='Set log level on ERROR',
                        action='store_true')
    parser.add_argument('-t', '--tokenize',
                        help='Token-based evaluation (tokenizer splits on whitespace)',
                        action='store_true')
    return parser.parse_args()


def main():
    args = parse_args()
    log_level = logging.WARNING
    if args.silent:
        log_level = logging.ERROR
    logging.basicConfig(level=log_level,
                        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

    token_func = None
    if args.tokenize:
        token_func = tokenize

    f1_agreement = compute_f1_agreement(args.project_root, token_func=token_func)
    iaa_report(f1_agreement, args.precision)
    if args.heatmap_path:
        f1_agreement.draw_heatmap(args.heatmap_path)


if __name__ == '__main__':
    main()



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='bratiaa',
      version='0.1.1',
      author='Tobias Kolditz',
      author_email='tbs.kldtz@gmail.com',
      description='Inter-annotator agreement for Brat annotation projects',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/kldtz/bratiaa',
      packages=['bratiaa', 'bratsubset'],
      install_requires=[
          'filelock',  # imported by bratsubset.annotation.py, but not used here (read-only)
          'numpy',
          'matplotlib',
          'pytest',
          'scipy',
          'tabulate'
      ],
      python_requires='>=3.5',
      entry_points={
          'console_scripts': [
              'brat-iaa=bratiaa.agree_cli:main'
          ]
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ]
      )



#!/usr/bin/env python
# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; coding: utf-8; -*-
# vim:set ft=python ts=4 sw=4 sts=4 autoindent:

"""Dummy Messager that can replace the real one in case it goes down. Doesn't
actually send any messages other than letting the user know of the problem. Use
e.g. as.

try:     from message import Messager except:     from sosmessage import
Messager
"""


class SosMessager:
    def output_json(json_dict):
        json_dict['messages'] = [
            ['HELP: messager down! (internal error in message.py, please contact administrator)', 'error', -1]]
        return json_dict
    output_json = staticmethod(output_json)

    def output(o):
        print('HELP: messager down! (internal error in message.py, please contact administrator)', file=o)
    output = staticmethod(output)

    def info(msg, duration=3, escaped=False): pass
    info = staticmethod(info)

    def warning(msg, duration=3, escaped=False): pass
    warning = staticmethod(warning)

    def error(msg, duration=3, escaped=False): pass
    error = staticmethod(error)

    def debug(msg, duration=3, escaped=False): pass
    debug = staticmethod(debug)



import re

import numpy as np

ENCODING = 'utf-8'

TOKEN = re.compile(r'\S+')


def tokenize(text):
    for match in re.finditer(TOKEN, text):
        yield match.start(), match.end()


def read(path, encoding=ENCODING, newline='\r\n', mode='r'):
    with open(path, newline=newline, encoding=encoding, mode=mode) as fin:
        return fin.read()


class TokenOverlap:
    """
    Data structure for quick lookup of tokens overlapping with given span.
    Assumes that the provided list of tokens is sorted by indices!
    """

    def __init__(self, text, tokens):
        self.tokens = tokens
        self.char2token = self.compute_mapping(len(text), tokens)

    @staticmethod
    def compute_mapping(text_length, tokens):
        char2token = np.zeros(text_length, dtype=int)
        i = 0
        for token_idx, (start, end) in enumerate(tokens):
            char2token[i:start] = token_idx - 1
            char2token[start:end] = token_idx
            i = end
        char2token[i:text_length] = len(tokens) - 1
        return char2token

    def overlapping_tokens(self, start, end):
        assert end <= len(self.char2token), f'End index {end} > text length {len(self.char2token)}!'
        if end < 1 or start >= end:
            return []
        start_token = self.char2token[start]
        if start_token == -1: # start offset before first token
            start_token = 0
        if self.tokens[start_token][1] <= start: # start offset between two tokens
            start_token += 1
        end_token = self.char2token[end - 1] # end offset is exclusive
        if end_token < 0 or end_token < start_token:
            return []
        return self.tokens[start_token:end_token + 1]



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; coding: utf-8; -*-

# This configuration file specifies the global setup of the brat
# server. It is recommended that you use the installation script
# instead of editing this file directly. To do this, run the following
# command in the brat directory:
#
#     ./install.sh
#
# if you wish to configure the server manually, you will first need to
# make sure that this file appears as config.py in the brat server
# root directory. If this file is currently named config.py,
# you can do this as follows:
#
#     cp config.py config.py
#
# you will then need to edit config.py, minimally replacing all
# instances of the string CHANGE_ME with their appropriate values.
# Please note that these values MUST appear in quotes, e.g. as in
#
# ADMIN_CONTACT_EMAIL = 'admin@example.com'


# Contact email for users to use if the software encounters errors
ADMIN_CONTACT_EMAIL = None

# Directories required by the brat server:
#
#     BASE_DIR: directory in which the server is installed
#     DATA_DIR: directory containing texts and annotations
#     WORK_DIR: directory that the server uses for temporary files
#
BASE_DIR = None
DATA_DIR = None
WORK_DIR = None

# If you have installed brat as suggested in the installation
# instructions, you can set up BASE_DIR, DATA_DIR and WORK_DIR by
# removing the three lines above and deleting the initial '#'
# character from the following four lines:

#from os.path import dirname, join
#BASE_DIR = dirname(__file__)
#DATA_DIR = join(BASE_DIR, 'data')
#WORK_DIR = join(BASE_DIR, 'work')

# To allow editing, include at least one USERNAME:PASSWORD pair below.
# The format is the following:
#
#     'USERNAME': 'PASSWORD',
#
# For example, user `editor` and password `annotate`:
#
#     'editor': 'annotate',

USER_PASSWORD = {
    #     (add USERNAME:PASSWORD pairs below this line.)
}


########## ADVANCED CONFIGURATION OPTIONS ##########

# The following options control advanced aspects of the brat server
# setup.  It is not necessary to edit these in a basic brat server
# installation.


# MAX_SEARCH_RESULT_NUMBER
# It may be a good idea to limit the max number of results to a search
# as very high numbers can be demanding of both server and clients.
# (unlimited if not defined or <= 0)

MAX_SEARCH_RESULT_NUMBER = 1000


# DEBUG
# Set to True to enable additional debug output

DEBUG = False

# TUTORIALS
# Unauthorised users can create tutorials (but not edit without a login)
TUTORIALS = False

# LOG_LEVEL
# If you are a developer you may want to turn on extensive server
# logging by enabling LOG_LEVEL = LL_DEBUG

LL_DEBUG, LL_INFO, LL_WARNING, LL_ERROR, LL_CRITICAL = list(range(5))
LOG_LEVEL = LL_WARNING
#LOG_LEVEL = LL_DEBUG

# BACKUP_DIR
# Define to enable backups

# from os.path import join
#BACKUP_DIR = join(WORK_DIR, 'backup')

#try:
#    assert DATA_DIR != BACKUP_DIR, 'DATA_DIR cannot equal BACKUP_DIR'
#except NameError:
#    pass  # BACKUP_DIR most likely not defined


# SVG_CONVERSION_COMMANDS
# If export to formats other than SVG is needed, the server must have
# a software capable of conversion like inkscape set up, and the
# following must be defined.
# (SETUP NOTE: at least Inkscape 0.46 requires the directory
# ".gnome2/" in the apache home directory and will crash if it doesn't
# exist.)

# SVG_CONVERSION_COMMANDS = [
#    ('png', 'inkscape --export-area-drawing --without-gui --file=%s --export-png=%s'),
#    ('pdf', 'inkscape --export-area-drawing --without-gui --file=%s --export-pdf=%s'),
#    ('eps', 'inkscape --export-area-drawing --without-gui --file=%s --export-eps=%s'),
# ]



#!/usr/bin/env python
# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; coding: utf-8; -*-
# vim:set ft=python ts=4 sw=4 sts=4 autoindent:

from __future__ import with_statement

from itertools import chain, takewhile
from os import close as os_close
from os.path import join as path_join
from os.path import splitext
from re import compile as re_compile
from re import match as re_match
from time import time

from filelock import FileLock

from bratsubset.common import ProtocolError
from bratsubset.message import Messager

'''
Functionality related to the annotation file format.

Author:     Pontus Stenetorp   <pontus is s u-tokyo ac jp>
Version:    2011-01-25
'''




# Constants
# The only suffix we allow to write to, which is the joined annotation file
JOINED_ANN_FILE_SUFF = 'ann'
# These file suffixes indicate partial annotations that can not be written to
# since they depend on multiple files for completeness
PARTIAL_ANN_FILE_SUFF = ['a1', 'a2', 'co', 'rel']
KNOWN_FILE_SUFF = [JOINED_ANN_FILE_SUFF] + PARTIAL_ANN_FILE_SUFF
TEXT_FILE_SUFFIX = 'txt'
# String used to catenate texts of discontinuous annotations in reference text
DISCONT_SEP = ' '
###

# If True, use BioNLP Shared Task 2013 compatibilty mode, allowing
# normalization annotations to be parsed using the BioNLP Shared Task
# 2013 format in addition to the bratsubset format and allowing relations to
# reference event triggers. NOTE: Alternate format supported only for
# reading normalization annotations, not for creating new ones. Don't
# change this setting unless you are sure you know what you are doing.
BIONLP_ST_2013_COMPATIBILITY = True
BIONLP_ST_2013_NORMALIZATION_RES = [
    (re_compile(r'^(Reference) Annotation:(\S+) Referent:(\S+)'), r'\1 \2 \3'),
    (re_compile(r'^(Reference) Referent:(\S+) Annotation:(\S+)'), r'\1 \3 \2'),
]


class AnnotationLineSyntaxError(Exception):
    def __init__(self, line, line_num, filepath):
        self.line = line
        self.line_num = line_num
        self.filepath = filepath

    def __str__(self):
        u'Syntax error on line %d: "%s"' % (self.line_num, self.line)


class IdedAnnotationLineSyntaxError(AnnotationLineSyntaxError):
    def __init__(self, id, line, line_num, filepath):
        AnnotationLineSyntaxError.__init__(self, line, line_num, filepath)
        self.id = id

    def __str__(self):
        u'Syntax error on line %d (id %s): "%s"' % (
            self.line_num, self.id, self.line)


class AnnotationNotFoundError(Exception):
    def __init__(self, id):
        self.id = id

    def __str__(self):
        return u'Could not find an annotation with id: %s' % (self.id, )


class AnnotationFileNotFoundError(ProtocolError):
    def __init__(self, fn):
        self.fn = fn

    def __str__(self):
        return u'Could not find any annotations for %s' % (self.fn, )

    def json(self, json_dic):
        json_dic['exception'] = 'annotationFileNotFound'
        return json_dic


class AnnotationCollectionNotFoundError(ProtocolError):
    def __init__(self, cn):
        self.cn = cn

    def __str__(self):
        return u'Error accessing collection %s' % (self.cn, )

    def json(self, json_dic):
        # TODO: more specific error?
        json_dic['exception'] = 'annotationCollectionNotFound'
        return json_dic


class EventWithoutTriggerError(ProtocolError):
    def __init__(self, event):
        self.event = event

    def __str__(self):
        return u'Event "%s" lacks a trigger' % (self.event, )

    def json(self, json_dic):
        json_dic['exception'] = 'eventWithoutTrigger'
        return json_dic


class EventWithNonTriggerError(ProtocolError):
    def __init__(self, event, non_trigger):
        self.event = event
        self.non_trigger = non_trigger

    def __str__(self):
        return u'Non-trigger "%s" used by "%s" as trigger' % (
            self.non_trigger, self.event, )

    def json(self, json_dic):
        json_dic['exception'] = 'eventWithNonTrigger'
        return json_dic


class TriggerReferenceError(ProtocolError):
    def __init__(self, trigger, referencer):
        self.trigger = trigger
        self.referencer = referencer

    def __str__(self):
        return u'Trigger "%s" referenced by non-event "%s"' % (
            self.trigger, self.referencer, )

    def json(self, json_dic):
        json_dic['exception'] = 'triggerReference'
        return json_dic


class AnnotationTextFileNotFoundError(AnnotationFileNotFoundError):
    def __str__(self):
        return u'Could not read text file for %s' % (self.fn, )


class AnnotationsIsReadOnlyError(ProtocolError):
    def __init__(self, fn):
        self.fn = fn

    def __str__(self):
        # No extra message; the client is doing a fine job of reporting this
        # return u'Annotations read-only for %s' % (self.fn, )
        return ''

    def json(self, json_dic):
        json_dic['exception'] = 'annotationIsReadOnly'
        return json_dic


class DuplicateAnnotationIdError(AnnotationLineSyntaxError):
    def __init__(self, id, line, line_num, filepath):
        AnnotationLineSyntaxError.__init__(self, line, line_num, filepath)
        self.id = id

    def __str__(self):
        return (u'Duplicate id: %s on line %d (id %s): "%s"'
                ) % (self.id, self.line_num, self.id, self.line)


class InvalidIdError(Exception):
    def __init__(self, id):
        self.id = id

    def __str__(self):
        return u'Invalid id: %s' % (self.id, )


class DependingAnnotationDeleteError(Exception):
    def __init__(self, target, dependants):
        self.target = target
        self.dependants = dependants

    def __str__(self):
        return u'%s can not be deleted due to depending annotations %s' % (str(
            self.target).rstrip(), ",".join([str(d).rstrip() for d in self.dependants]))

    def html_error_str(self, response=None):
        return u'''Annotation:
        %s
        Has depending annotations attached to it:
        %s''' % (str(self.target).rstrip(), ",".join([str(d).rstrip() for d in self.dependants]))


class SpanOffsetOverlapError(ProtocolError):
    def __init__(self, offsets):
        self.offsets = offsets

    def __str__(self):
        return u'The offsets [%s] overlap' % (
            ', '.join(str(e) for e in self.offsets))

    def json(self, json_dic):
        json_dic['exception'] = 'spanOffsetOverlapError'
        return json_dic


# Open function that enforces strict, utf-8, and universal newlines for reading
# TODO: Could have another wrapping layer raising an appropriate ProtocolError
def open_textfile(filename, mode='r'):
    return open(filename, mode, encoding='utf8', errors='strict')


def __split_annotation_id(id):
    m = re_match(r'^([A-Za-z]+|#[A-Za-z]*)([0-9]+)(.*?)$', id)
    if m is None:
        raise InvalidIdError(id)
    pre, num_str, suf = m.groups()
    return pre, num_str, suf


def annotation_id_prefix(id):
    pre = ''.join(c for c in takewhile(lambda x: not x.isdigit(), id))
    if not pre:
        raise InvalidIdError(id)
    return pre


def annotation_id_number(id):
    return __split_annotation_id(id)[1]


def is_valid_id(id):
    # special case: '*' is acceptable as an "ID"
    if id == '*':
        return True

    try:
        # currently accepting any ID that can be split.
        # TODO: consider further constraints
        __split_annotation_id(id)[1]
        return True
    except InvalidIdError:
        return False


class Annotations(object):
    """Basic annotation storage.

    Not concerned with conformity of annotations to text; can be created
    without access to the text file to which the annotations apply.
    """

    def get_document(self):
        return self._document

    def _select_input_files(self, document):
        """Given a document name (path), returns a list of the names of
        specific annotation files relevant to the document, or the empty list
        if none found.

        For example, given "1000", may return ["1000.a1", "1000.a2"].
        May set self._read_only flag to True.
        """

        from os.path import isfile
        from os import access, W_OK

        try:
            # Do we have a valid suffix? If so, it is probably best to the file
            suff = document[document.rindex('.') + 1:]
            if suff == JOINED_ANN_FILE_SUFF:
                # It is a joined file, let's load it
                input_files = [document]
                # Do we lack write permissions?
                if not access(document, W_OK):
                    # TODO: Should raise an exception or warning
                    self._read_only = True
            elif suff in PARTIAL_ANN_FILE_SUFF:
                # It is only a partial annotation, we will most likely fail
                # but we will try opening it
                input_files = [document]
                self._read_only = True
            else:
                input_files = []
        except ValueError:
            # The document lacked a suffix
            input_files = []

        if not input_files:
            # Our first attempts at finding the input by checking suffixes
            # failed, so we try to attach know suffixes to the path.
            sugg_path = document + '.' + JOINED_ANN_FILE_SUFF
            if isfile(sugg_path):
                # We found a joined file by adding the joined suffix
                input_files = [sugg_path]
                # Do we lack write permissions?
                if not access(sugg_path, W_OK):
                    # TODO: Should raise an exception or warning
                    self._read_only = True
            else:
                # Our last shot, we go for as many partial files as possible
                input_files = [sugg_path for sugg_path in
                               (document + '.' + suff
                                for suff in PARTIAL_ANN_FILE_SUFF)
                               if isfile(sugg_path)]
                self._read_only = True

        return input_files

    # TODO: DOC!
    def __init__(self, document, read_only=False):
        # this decides which parsing function is invoked by annotation
        # ID prefix (first letter)
        self._parse_function_by_id_prefix = {
            'T': self._parse_textbound_annotation,
            'M': self._parse_modifier_annotation,
            'A': self._parse_attribute_annotation,
            'N': self._parse_normalization_annotation,
            'R': self._parse_relation_annotation,
            '*': self._parse_equiv_annotation,
            'E': self._parse_event_annotation,
            '#': self._parse_comment_annotation,
        }

        # TODO: DOC!
        # TODO: Incorparate file locking! Is the destructor called upon inter
        # crash?
        from collections import defaultdict
        from os.path import getctime, getmtime
        #from fileinput import FileInput, hook_encoded

        # we should remember this
        self._document = document

        self.failed_lines = []
        self.externally_referenced_triggers = set()

        # Here be dragons, these objects need constant updating and syncing
        # Annotation for each line of the file
        self._lines = []
        # Mapping between annotation objects and which line they occur on
        # Range: [0, inf.) unlike [1, inf.) which is common for files
        self._line_by_ann = {}
        # Maximum id number used for each id prefix, to speed up id generation
        # XXX: This is effectively broken by the introduction of id suffixes
        self._max_id_num_by_prefix = defaultdict(lambda: 1)
        # Annotation by id, not includid non-ided annotations
        self._ann_by_id = {}
        ###

        # We use some heuristics to find the appropriate annotation files
        self._read_only = read_only
        input_files = self._select_input_files(document)

        if not input_files:
            with open('{}.{}'.format(document, JOINED_ANN_FILE_SUFF), 'w'):
                pass

            input_files = self._select_input_files(document)
            if not input_files:
                raise AnnotationFileNotFoundError(document)

        # We then try to open the files we got using the heuristics
        # self._file_input = FileInput(openhook=hook_encoded('utf-8'))
        self._input_files = input_files

        # Finally, parse the given annotation file
        self._parse_ann_file()

        # Sanity checking that can only be done post-parse
        self._sanity()
        # XXX: Hack to get the timestamps after parsing
        if (len(self._input_files) == 1 and
                self._input_files[0].endswith(JOINED_ANN_FILE_SUFF)):
            self.ann_mtime = getmtime(self._input_files[0])
            self.ann_ctime = getctime(self._input_files[0])
        else:
            # We don't have a single file, just set to epoch for now
            self.ann_mtime = -1
            self.ann_ctime = -1

    def _sanity(self):
        # Beware, we ONLY do format checking, leave your semantics hat at home

        # Check that referenced IDs are defined
        for ann in self:
            for rid in chain(*ann.get_deps()):
                try:
                    self.get_ann_by_id(rid)
                except AnnotationNotFoundError:
                    # TODO: do more than just send a message for this error?
                    Messager.error(
                        'ID ' +
                        rid +
                        ' not defined, referenced from annotation ' +
                        str(ann))

        # Check that each event has a trigger
        for e_ann in self.get_events():
            try:
                tr_ann = self.get_ann_by_id(e_ann.trigger)

                # If the annotation is not text-bound or of different type
                if (not isinstance(tr_ann, TextBoundAnnotation) or
                        tr_ann.type != e_ann.type):
                    raise EventWithNonTriggerError(e_ann, tr_ann)
            except AnnotationNotFoundError:
                raise EventWithoutTriggerError(e_ann)

        # Check that every trigger is only referenced by events

        # Create a map for non-event references
        referenced_to_referencer = {}
        for non_e_ann in (a for a in self
                          if not isinstance(a, EventAnnotation)
                          and isinstance(a, IdedAnnotation)):
            for ref in chain(*non_e_ann.get_deps()):
                try:
                    referenced_to_referencer[ref].add(non_e_ann.id)
                except KeyError:
                    referenced_to_referencer[ref] = set((non_e_ann.id, ))

        # Ensure that no non-event references a trigger
        for tr_ann in self.get_triggers():
            if tr_ann.id in referenced_to_referencer:
                conflict_ann_ids = referenced_to_referencer[tr_ann.id]
                if BIONLP_ST_2013_COMPATIBILITY:
                    # Special-case processing for BioNLP ST 2013: allow
                    # Relations to reference event triggers (#926).
                    remaining_confict_ann_ids = set()
                    for rid in conflict_ann_ids:
                        referencer = self.get_ann_by_id(rid)
                        if not isinstance(
                                referencer, BinaryRelationAnnotation):
                            remaining_confict_ann_ids.add(rid)
                        else:
                            self.externally_referenced_triggers.add(tr_ann.id)
                    conflict_ann_ids = remaining_confict_ann_ids
                # Note: Only reporting one of the conflicts (TODO)
                if conflict_ann_ids:
                    referencer = self.get_ann_by_id(list(conflict_ann_ids)[0])
                    raise TriggerReferenceError(tr_ann, referencer)

    def get_events(self):
        return (a for a in self if isinstance(a, EventAnnotation))

    def get_attributes(self):
        return (a for a in self if isinstance(a, AttributeAnnotation))

    def get_equivs(self):
        return (a for a in self if isinstance(a, EquivAnnotation))

    def get_textbounds(self):
        return (a for a in self if isinstance(a, TextBoundAnnotation))

    def get_relations(self):
        return (a for a in self if isinstance(a, BinaryRelationAnnotation))

    def get_normalizations(self):
        return (a for a in self if isinstance(a, NormalizationAnnotation))

    def get_entities(self):
        # Entities are textbounds that are not triggers
        triggers = [t for t in self.get_triggers()]
        return (a for a in self if (isinstance(a, TextBoundAnnotation) and
                                    a not in triggers))

    def get_oneline_comments(self):
        # XXX: The status exception is for the document status protocol
        #       which is yet to be formalised
        return (a for a in self if isinstance(a, OnelineCommentAnnotation)
                and a.type != 'STATUS')

    def get_statuses(self):
        return (a for a in self if isinstance(a, OnelineCommentAnnotation)
                and a.type == 'STATUS')

    def get_triggers(self):
        # Triggers are text-bounds referenced by events
        # TODO: this omits entity triggers that lack a referencing event
        # (for one reason or another -- bratsubset shouldn't define any.)
        return (self.get_ann_by_id(e.trigger) for e in self.get_events())

    # TODO: getters for other categories of annotations
    # TODO: Remove read and use an internal and external version instead
    def add_annotation(self, ann, read=False):
        # log_info(u'Will add: ' + str(ann).rstrip('\n') + ' ' + str(type(ann)))
        # TODO: DOC!
        # TODO: Check read only
        if not read and self._read_only:
            raise AnnotationsIsReadOnlyError(self.get_document())

        # Equivs have to be merged with other equivs
        try:
            # Bail as soon as possible for non-equivs
            ann.entities  # TODO: what is this?
            merge_cand = ann
            for eq_ann in self.get_equivs():
                try:
                    # Make sure that this Equiv duck quacks
                    eq_ann.entities
                except AttributeError as e:
                    assert False, 'got a non-entity from an entity call'

                # Do we have an entitiy in common with this equiv?
                for ent in merge_cand.entities:
                    if ent in eq_ann.entities:
                        for m_ent in merge_cand.entities:
                            if m_ent not in eq_ann.entities:
                                eq_ann.entities.append(m_ent)
                        # Don't try to delete ann since it never was added
                        if merge_cand != ann:
                            try:
                                self.del_annotation(merge_cand)
                            except DependingAnnotationDeleteError:
                                assert False, ('Equivs lack ids and should '
                                               'never have dependent annotations')
                        merge_cand = eq_ann
                        # We already merged it all, break to the next ann
                        break

            if merge_cand != ann:
                # The proposed annotation was simply merged, no need to add it
                # Update the modification time
                from time import time
                self.ann_mtime = time()
                return

        except AttributeError:
            # XXX: This can catch a ton more than we want to! Ugly!
            # It was not an Equiv, skip along
            pass

        # Register the object id
        try:
            self._ann_by_id[ann.id] = ann
            pre, num = annotation_id_prefix(
                ann.id), annotation_id_number(
                ann.id)
            self._max_id_num_by_prefix[pre] = max(
                int(num), int(self._max_id_num_by_prefix[pre]))
        except AttributeError:
            # The annotation simply lacked an id which is fine
            pass

        # Add the annotation as the last line
        self._lines.append(ann)
        self._line_by_ann[ann] = len(self) - 1
        # Update the modification time
        from time import time
        self.ann_mtime = time()

    def del_annotation(self, ann, tracker=None):
        # TODO: Check read only
        # TODO: Flag to allow recursion
        # TODO: Sampo wants to allow delet of direct deps but not indirect, one step
        # TODO: needed to pass tracker to track recursive mods, but use is too
        #      invasive (direct modification of ModificationTracker.deleted)
        # TODO: DOC!
        if self._read_only:
            raise AnnotationsIsReadOnlyError(self.get_document())

        try:
            ann.id
        except AttributeError:
            # If it doesn't have an id, nothing can depend on it
            if tracker is not None:
                tracker.deletion(ann)
            self._atomic_del_annotation(ann)
            # Update the modification time
            from time import time
            self.ann_mtime = time()
            return

        # collect annotations dependending on ann
        ann_deps = []

        for other_ann in self:
            soft_deps, hard_deps = other_ann.get_deps()
            if str(ann.id) in soft_deps | hard_deps:
                ann_deps.append(other_ann)

        # If all depending are AttributeAnnotations or EquivAnnotations,
        # delete all modifiers recursively (without confirmation) and remove
        # the annotation id from the equivs (and remove the equiv if there is
        # only one id left in the equiv)
        # Note: this assumes AttributeAnnotations cannot have
        # other dependencies depending on them, nor can EquivAnnotations
        if all((False for d in ann_deps if (
            not isinstance(d, AttributeAnnotation)
            and not isinstance(d, EquivAnnotation)
            and not isinstance(d, OnelineCommentAnnotation)
            and not isinstance(d, NormalizationAnnotation)
        ))):

            for d in ann_deps:
                if isinstance(d, AttributeAnnotation):
                    if tracker is not None:
                        tracker.deletion(d)
                    self._atomic_del_annotation(d)
                elif isinstance(d, EquivAnnotation):
                    if len(d.entities) <= 2:
                        # An equiv has to have more than one member
                        self._atomic_del_annotation(d)
                        if tracker is not None:
                            tracker.deletion(d)
                    else:
                        if tracker is not None:
                            before = str(d)
                        d.entities.remove(str(ann.id))
                        if tracker is not None:
                            tracker.change(before, d)
                elif isinstance(d, OnelineCommentAnnotation):
                    # TODO: Can't anything refer to comments?
                    self._atomic_del_annotation(d)
                    if tracker is not None:
                        tracker.deletion(d)
                elif isinstance(d, NormalizationAnnotation):
                    # Nothing should be able to reference normalizations
                    self._atomic_del_annotation(d)
                    if tracker is not None:
                        tracker.deletion(d)
                else:
                    # all types we allow to be deleted along with
                    # annotations they depend on should have been
                    # covered above.
                    assert False, "INTERNAL ERROR"
            ann_deps = []

        if ann_deps:
            raise DependingAnnotationDeleteError(ann, ann_deps)

        if tracker is not None:
            tracker.deletion(ann)
        self._atomic_del_annotation(ann)

    def _atomic_del_annotation(self, ann):
        # TODO: DOC
        # Erase the ann by id shorthand
        try:
            del self._ann_by_id[ann.id]
        except AttributeError:
            # So, we did not have id to erase in the first place
            pass

        ann_line = self._line_by_ann[ann]
        # Erase the main annotation
        del self._lines[ann_line]
        # Erase the ann by line shorthand
        del self._line_by_ann[ann]
        # Update the line shorthand of every annotation after this one
        # to reflect the new self._lines
        for l_num in range(ann_line, len(self)):
            self._line_by_ann[self[l_num]] = l_num
        # Update the modification time
        from time import time
        self.ann_mtime = time()

    def get_ann_by_id(self, id):
        # TODO: DOC
        try:
            return self._ann_by_id[id]
        except KeyError:
            raise AnnotationNotFoundError(id)

    def get_new_id(self, prefix, suffix=None):
        """Return a new valid unique id for this annotation file for the given
        prefix. No ids are re-used for traceability over time for annotations,
        but this only holds for the lifetime of the annotation object. If the
        annotation file is parsed once again into an annotation object the next
        assigned id will be the maximum seen for a given prefix plus one which
        could have been deleted during a previous annotation session.

        Warning: get_new_id('T') == get_new_id('T')
        Just calling this method does not reserve the id, you need to
        add the annotation with the returned id to the annotation object in
        order to reserve it.

        Argument(s):
        id_pre - an annotation prefix on the format [A-Za-z]+

        Returns:
        An id that is guaranteed to be unique for the lifetime of the
        annotation.
        """
        # XXX: We have changed this one radically!
        # XXX: Stupid and linear
        if suffix is None:
            suffix = ''
        # XXX: Arbitrary constant!
        for suggestion in (
            prefix +
            str(i) +
            suffix for i in range(
                1,
                2**15)):
            # This is getting more complicated by the minute, two checks since
            # the developers no longer know when it is an id or string.
            if suggestion not in self._ann_by_id:
                return suggestion

    # XXX: This syntax is subject to change
    def _parse_attribute_annotation(
            self, id, data, data_tail, input_file_path):
        match = re_match(r'(.+?) (.+?) (.+?)$', data)
        if match is None:
            # Is it an old format without value?
            match = re_match(r'(.+?) (.+?)$', data)

            if match is None:
                raise IdedAnnotationLineSyntaxError(
                    id, self.ann_line, self.ann_line_num + 1, input_file_path)

            _type, target = match.groups()
            value = True
        else:
            _type, target, value = match.groups()

        # Verify that the ID is indeed valid
        try:
            annotation_id_number(target)
        except InvalidIdError:
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        return AttributeAnnotation(
            target, id, _type, '', value, source_id=input_file_path)

    def _parse_event_annotation(self, id, data, data_tail, input_file_path):
        # XXX: A bit nasty, we require a single space
        try:
            type_delim = data.index(' ')
            type_trigger, type_trigger_tail = (
                data[:type_delim], data[type_delim:])
        except ValueError:
            type_trigger = data.rstrip('\r\n')
            type_trigger_tail = None

        try:
            type, trigger = type_trigger.split(':')
        except ValueError:
            # TODO: consider accepting events without triggers, e.g.
            # BioNLP ST 2011 Bacteria task
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        if type_trigger_tail is not None:
            args = [tuple(arg.split(':')) for arg in type_trigger_tail.split()]
        else:
            args = []

        return EventAnnotation(
            trigger,
            args,
            id,
            type,
            data_tail,
            source_id=input_file_path)

    def _parse_relation_annotation(self, id, data, data_tail, input_file_path):
        try:
            type_delim = data.index(' ')
            type, type_tail = (data[:type_delim], data[type_delim:])
        except ValueError:
            # cannot have a relation with just a type (contra event)
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        try:
            args = [tuple(arg.split(':')) for arg in type_tail.split()]
        except ValueError:
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        if len(args) != 2:
            Messager.error(
                'Error parsing relation: must have exactly two arguments')
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        if args[0][0] == args[1][0]:
            Messager.error(
                'Error parsing relation: arguments must not be identical')
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        return BinaryRelationAnnotation(id, type,
                                        args[0][0], args[0][1],
                                        args[1][0], args[1][1],
                                        data_tail, source_id=input_file_path)

    def _parse_equiv_annotation(self, dummy, data, data_tail, input_file_path):
        # NOTE: first dummy argument to have a uniform signature with other
        # parse_* functions
        # TODO: this will split on any space, which is likely not correct
        try:
            type, type_tail = data.split(None, 1)
        except ValueError:
            # no space: Equiv without arguments?
            raise AnnotationLineSyntaxError(
                self.ann_line, self.ann_line_num + 1, input_file_path)
        equivs = type_tail.split(None)
        return EquivAnnotation(type, equivs, data_tail,
                               source_id=input_file_path)

    # Parse an old modifier annotation for backwards compatibility
    def _parse_modifier_annotation(self, id, data, data_tail, input_file_path):
        type, target = data.split()
        return AttributeAnnotation(
            target,
            id,
            type,
            data_tail,
            True,
            source_id=input_file_path)

    def _split_textbound_data(self, id, data, input_file_path):
        try:
            # first space-separated string is type
            type, rest = data.split(' ', 1)

            # rest should be semicolon-separated list of "START END"
            # pairs, where START and END are integers
            spans = []
            for span_str in rest.split(';'):
                start_str, end_str = span_str.split(' ', 2)

                # ignore trailing whitespace
                end_str = end_str.rstrip()

                if any((c.isspace() for c in end_str)):
                    Messager.error(
                        'Error parsing textbound "%s\t%s". (Using space instead of tab?)' %
                        (id, data))
                    raise IdedAnnotationLineSyntaxError(
                        id, self.ann_line, self.ann_line_num + 1, input_file_path)

                start, end = (int(start_str), int(end_str))
                spans.append((start, end))

        except BaseException:
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        return type, spans

    def _parse_textbound_annotation(
            self, _id, data, data_tail, input_file_path):
        _type, spans = self._split_textbound_data(_id, data, input_file_path)
        return TextBoundAnnotation(
            spans, _id, _type, data_tail, source_id=input_file_path)

    def _parse_normalization_annotation(
            self, _id, data, data_tail, input_file_path):
        # special-case processing for BioNLP ST 2013 variant of
        # normalization format
        if BIONLP_ST_2013_COMPATIBILITY:
            for r, s in BIONLP_ST_2013_NORMALIZATION_RES:
                d = r.sub(s, data, count=1)
                if d != data:
                    data = d
                    break

        match = re_match(r'(\S+) (\S+) (\S+?):(\S+)', data)
        if match is None:
            raise IdedAnnotationLineSyntaxError(
                _id, self.ann_line, self.ann_line_num + 1, input_file_path)
        _type, target, refdb, refid = match.groups()

        return NormalizationAnnotation(
            _id,
            _type,
            target,
            refdb,
            refid,
            data_tail,
            source_id=input_file_path)

    def _parse_comment_annotation(self, _id, data, data_tail, input_file_path):
        try:
            _type, target = data.split()
        except ValueError:
            raise IdedAnnotationLineSyntaxError(
                _id, self.ann_line, self.ann_line_num + 1, input_file_path)
        return OnelineCommentAnnotation(
            target, _id, _type, data_tail, source_id=input_file_path)

    def _parse_ann_file(self):
        self.ann_line_num = -1
        for input_file_path in self._input_files:
            with open_textfile(input_file_path) as input_file:
                # for self.ann_line_num, self.ann_line in
                # enumerate(self._file_input):
                for self.ann_line in input_file:
                    self.ann_line_num += 1
                    try:
                        # ID processing
                        try:
                            id, id_tail = self.ann_line.split('\t', 1)
                        except ValueError:
                            raise AnnotationLineSyntaxError(
                                self.ann_line, self.ann_line_num + 1, input_file_path)

                        pre = annotation_id_prefix(id)

                        if id in self._ann_by_id and pre != '*':
                            raise DuplicateAnnotationIdError(
                                id, self.ann_line, self.ann_line_num + 1, input_file_path)

                        # if the ID is not valid, need to fail with
                        # AnnotationLineSyntaxError (not
                        # IdedAnnotationLineSyntaxError).
                        if not is_valid_id(id):
                            raise AnnotationLineSyntaxError(
                                self.ann_line, self.ann_line_num + 1, input_file_path)

                        # Cases for lines
                        try:
                            data_delim = id_tail.index('\t')
                            data, data_tail = (id_tail[:data_delim],
                                               id_tail[data_delim:])
                        except ValueError:
                            data = id_tail
                            # No tail at all, although it should have a \t
                            data_tail = ''

                        new_ann = None

                        #log_info('Will evaluate prefix: ' + pre)

                        assert len(pre) >= 1, "INTERNAL ERROR"
                        pre_first = pre[0]

                        try:
                            parse_func = self._parse_function_by_id_prefix[pre_first]
                            new_ann = parse_func(
                                id, data, data_tail, input_file_path)
                        except KeyError:
                            raise IdedAnnotationLineSyntaxError(
                                id, self.ann_line, self.ann_line_num + 1, input_file_path)

                        assert new_ann is not None, "INTERNAL ERROR"
                        self.add_annotation(new_ann, read=True)
                    except IdedAnnotationLineSyntaxError as e:
                        # Could parse an ID but not the whole line; add
                        # UnparsedIdedAnnotation
                        self.add_annotation(
                            UnparsedIdedAnnotation(
                                e.id, e.line, source_id=e.filepath), read=True)
                        self.failed_lines.append(e.line_num - 1)

                    except AnnotationLineSyntaxError as e:
                        # We could not parse even an ID on the line, just add
                        # it as an unknown annotation
                        self.add_annotation(
                            UnknownAnnotation(
                                e.line,
                                source_id=e.filepath),
                            read=True)
                        # NOTE: For access we start at line 0, not 1 as in here
                        self.failed_lines.append(e.line_num - 1)

    def __str__(self):
        s = u'\n'.join(str(ann).rstrip(u'\r\n') for ann in self)
        if not s:
            return u''
        else:
            return s if s[-1] == u'\n' else s + u'\n'

    def __it__(self):
        for ann in self._lines:
            yield ann

    def __getitem__(self, val):
        try:
            # First, try to use it as a slice object
            return self._lines[val.start, val.stop, val.step]
        except AttributeError:
            # It appears not to be a slice object, try it as an index
            return self._lines[val]

    def __len__(self):
        return len(self._lines)

    def __enter__(self):
        # No need to do any handling here, the constructor handles that
        return self

    def __exit__(self, type, value, traceback):
        # self._file_input.close()
        if not self._read_only:
            assert len(self._input_files) == 1, 'more than one valid outfile'

            # We are hitting the disk a lot more than we should here, what we
            # should have is a modification flag in the object but we can't
            # due to how we change the annotations.

            out_str = str(self)
            with open_textfile(self._input_files[0], 'r') as old_ann_file:
                old_str = old_ann_file.read()

            # Was it changed?
            if out_str == old_str:
                # Then just return
                return

            from bratsubset.config import WORK_DIR

            # Protect the write so we don't corrupt the file
            with FileLock(path_join(
                    WORK_DIR, str(hash(self._input_files[0].replace('/', '_'))) + '.lock'
            )) as lock_file:
                #from tempfile import NamedTemporaryFile
                from tempfile import mkstemp
                # TODO: XXX: Is copyfile really atomic?
                from shutil import copyfile
                # XXX: NamedTemporaryFile only supports encoding for Python 3
                #       so we hack around it.
                # with NamedTemporaryFile('w', suffix='.ann') as tmp_file:
                # Grab the filename, but discard the handle
                tmp_fh, tmp_fname = mkstemp(suffix='.ann')
                os_close(tmp_fh)
                try:
                    with open_textfile(tmp_fname, 'w') as tmp_file:
                        # XXX: Temporary hack to make sure we don't write corrupted
                        #       files, but the client will already have the version
                        #       at this stage leading to potential problems upon
                        #       the next change to the file.
                        tmp_file.write(out_str)
                        tmp_file.flush()

                        try:
                            with Annotations(tmp_file.name) as ann:
                                # Move the temporary file onto the old file
                                copyfile(tmp_file.name, self._input_files[0])
                                # As a matter of convention we adjust the modified
                                # time of the data dir when we write to it. This
                                # helps us to make back-ups
                                time()
                                # XXX: Disabled for now!
                                #utime(DATA_DIR, (now, now))
                        except Exception as e:
                            Messager.error(
                                'ERROR writing changes: generated annotations cannot be read back in!\n(This is almost certainly a system error, please contact the developers.)\n%s' %
                                e, -1)
                            raise
                finally:
                    try:
                        from os import remove
                        remove(tmp_fname)
                    except Exception as e:
                        Messager.error(
                            "Error removing temporary file '%s'" %
                            tmp_fname)
            return

    def __in__(self, other):
        # XXX: You should do this one!
        pass


class TextAnnotations(Annotations):
    """Text-bound annotation storage.

    Extends Annotations in assuming access to text text to which the
    annotations apply and verifying the correctness of text-bound
    annotations against the text.
    """

    def __init__(self, document, read_only=False):
        # First read the text or the Annotations can't verify the annotations
        if document.endswith('.txt'):
            textfile_path = document
        else:
            # Do we have a known extension?
            _, file_ext = splitext(document)
            if not file_ext or file_ext not in KNOWN_FILE_SUFF:
                textfile_path = document
            else:
                textfile_path = document[:len(document) - len(file_ext)]

        self._document_text = self._read_document_text(textfile_path)

        Annotations.__init__(self, document, read_only)

    def _parse_textbound_annotation(
            self, id, data, data_tail, input_file_path):
        type, spans = self._split_textbound_data(id, data, input_file_path)

        # Verify spans
        seen_spans = []
        for start, end in spans:
            if start > end:
                Messager.error('Text-bound annotation start > end.')
                raise IdedAnnotationLineSyntaxError(
                    id, self.ann_line, self.ann_line_num + 1, input_file_path)
            if start < 0:
                Messager.error('Text-bound annotation start < 0.')
                raise IdedAnnotationLineSyntaxError(
                    id, self.ann_line, self.ann_line_num + 1, input_file_path)
            if end > len(self._document_text):
                Messager.error(
                    'Text-bound annotation offset exceeds text length.')
                raise IdedAnnotationLineSyntaxError(
                    id, self.ann_line, self.ann_line_num + 1, input_file_path)

            for ostart, oend in seen_spans:
                if end >= ostart and start < oend:
                    Messager.error('Text-bound annotation spans overlap')
                    raise IdedAnnotationLineSyntaxError(
                        id, self.ann_line, self.ann_line_num + 1, input_file_path)

            seen_spans.append((start, end))

        # first part is text, second connecting separators
        spanlen = sum([end - start for start, end in spans]) + \
            (len(spans) - 1) * len(DISCONT_SEP)

        # Require tail to be either empty or to begin with the text
        # corresponding to the catenation of the start:end spans.
        # If the tail is empty, force a fill with the corresponding text.
        if data_tail.strip() == '' and spanlen > 0:
            Messager.error(
                u"Text-bound annotation missing text (expected format 'ID\\tTYPE START END\\tTEXT'). Filling from reference text. NOTE: This changes annotations on disk unless read-only.")
            text = "".join([self._document_text[start:end]
                            for start, end in spans])

        elif data_tail[0] != '\t':
            Messager.error(
                'Text-bound annotation missing tab before text (expected format "ID\\tTYPE START END\\tTEXT").')
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        elif spanlen > len(data_tail) - 1:  # -1 for tab
            Messager.error(
                'Text-bound annotation text "%s" shorter than marked span(s) %s' % (data_tail[1:], str(spans)))
            raise IdedAnnotationLineSyntaxError(
                id, self.ann_line, self.ann_line_num + 1, input_file_path)

        else:
            text = data_tail[1:spanlen + 1]  # shift 1 for tab
            data_tail = data_tail[spanlen + 1:]

            spantexts = [self._document_text[start:end]
                         for start, end in spans]
            reftext = DISCONT_SEP.join(spantexts)

            if text != reftext:
                # just in case someone has been running an old version of
                # discont that catenated spans without DISCONT_SEP
                oldstylereftext = ''.join(spantexts)
                if text[:len(oldstylereftext)] == oldstylereftext:
                    Messager.warning(
                        u'NOTE: replacing old-style (pre-1.3) discontinuous annotation text span with new-style one, i.e. adding space to "%s" in .ann' % text[:len(oldstylereftext)], -1)
                    text = reftext
                    data_tail = ''
                else:
                    # unanticipated mismatch
                    Messager.error(
                        (u'Text-bound annotation text "%s" does not '
                         u'match marked span(s) %s text "%s" in document') %
                        (text, str(spans), reftext.replace(
                            '\n', '\\n')))
                    raise IdedAnnotationLineSyntaxError(
                        id, self.ann_line, self.ann_line_num + 1, input_file_path)

            if data_tail != '' and not data_tail[0].isspace():
                Messager.error(
                    u'Text-bound annotation text "%s" not separated from rest of line ("%s") by space!' %
                    (text, data_tail))
                raise IdedAnnotationLineSyntaxError(
                    id, self.ann_line, self.ann_line_num + 1, input_file_path)

        return TextBoundAnnotationWithText(
            spans, id, type, text, data_tail, source_id=input_file_path)

    def get_document_text(self):
        return self._document_text

    def _read_document_text(self, document):
        # TODO: this is too naive; document may be e.g. "PMID.a1",
        # in which case the reasonable text file name guess is
        # "PMID.txt", not "PMID.a1.txt"
        textfn = document + '.' + TEXT_FILE_SUFFIX
        try:
            with open_textfile(textfn, 'r') as f:
                return f.read()
        except IOError:
            Messager.error('Error reading document text from %s' % textfn)
        raise AnnotationTextFileNotFoundError(document)


class Annotation(object):
    """Base class for all annotations."""

    def __init__(self, tail, source_id=None):
        self.tail = tail
        self.source_id = source_id

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return u'%s("%s")' % (str(self.__class__), str(self))

    def get_deps(self):
        return (set(), set())


class UnknownAnnotation(Annotation):
    """Represents a line of annotation that could not be parsed.

    These are not discarded, but rather passed through unmodified.
    """

    def __init__(self, line, source_id=None):
        Annotation.__init__(self, line, source_id=source_id)

    def __str__(self):
        return self.tail


class UnparsedIdedAnnotation(Annotation):
    """Represents an annotation for which an ID could be read but the rest of
    the line could not be parsed.

    This is separate from UnknownAnnotation to allow IDs for unparsed
    annotations to be "reserved".
    """
    # duck-type instead of inheriting from IdedAnnotation as
    # that inherits from TypedAnnotation and we have no type

    def __init__(self, id, line, source_id=None):
        # (this actually is the whole line, not just the id tail,
        # although Annotation will assign it to self.tail)
        Annotation.__init__(self, line, source_id=source_id)
        self.id = id

    def __str__(self):
        return str(self.tail)


class TypedAnnotation(Annotation):
    """Base class for all annotations with a type."""

    def __init__(self, type, tail, source_id=None):
        Annotation.__init__(self, tail, source_id=source_id)
        self.type = type

    def __str__(self):
        raise NotImplementedError


class IdedAnnotation(TypedAnnotation):
    """Base class for all annotations with an ID."""

    def __init__(self, id, type, tail, source_id=None):
        TypedAnnotation.__init__(self, type, tail, source_id=source_id)
        self.id = id

    def reference_id(self):
        """Returns a list that uniquely identifies this annotation within its
        document."""
        return [self.id]

    def reference_text(self):
        """Returns a human-readable string that identifies this annotation
        within its document."""
        return str(self.reference_id()[0])

    def __str__(self):
        raise NotImplementedError


def split_role(r):
    """Given a string R that may be suffixed with a number, returns a tuple
    (ROLE, NUM) where ROLE+NUM == R and NUM is the maximal suffix of R
    consisting only of digits."""
    i = len(r)
    while i > 1 and r[i - 1].isdigit():
        i -= 1
    return r[:i], r[i:]


class EventAnnotation(IdedAnnotation):
    """Represents an event annotation. Events are typed annotations that are
    associated with a specific text expression stating the event (TRIGGER,
    identifying a TextBoundAnnotation) and have an arbitrary number of
    arguments, each of which is represented as a ROLE:PARTID pair, where ROLE
    is a string identifying the role (e.g. "Theme", "Cause") and PARTID the ID
    of another annotation participating in the event.

    Represented in standoff as

    ID\tTYPE:TRIGGER [ROLE1:PART1 ROLE2:PART2 ...]
    """

    def __init__(self, trigger, args, id, type, tail, source_id=None):
        IdedAnnotation.__init__(self, id, type, tail, source_id=source_id)
        self.trigger = trigger
        self.args = args

    def add_argument(self, role, argid):
        # split into "main" role label and possible numeric suffix
        role, rnum = split_role(role)
        if rnum != '':
            # if given a role with an explicit numeric suffix,
            # use the role as given (assume number is part of
            # role label).
            pass
        else:
            # find next free numeric suffix.

            # for each argument role in existing roles, determine the
            # role numbers already used
            rnums = {}
            for r, aid in self.args:
                rb, rn = split_role(r)
                if rb not in rnums:
                    rnums[rb] = {}
                rnums[rb][rn] = True

            # find the first available free number for the current role,
            # using the convention that the empty number suffix stands for 1
            rnum = ''
            while role in rnums and rnum in rnums[role]:
                if rnum == '':
                    rnum = '2'
                else:
                    rnum = str(int(rnum) + 1)

        # role+rnum is available, add
        self.args.append((role + rnum, argid))

    def __str__(self):
        return u'%s\t%s:%s %s%s' % (
            self.id,
            self.type,
            self.trigger,
            ' '.join([':'.join(map(str, arg_tup))
                      for arg_tup in self.args]),
            self.tail
        )

    def get_deps(self):
        soft_deps, hard_deps = IdedAnnotation.get_deps(self)
        hard_deps.add(self.trigger)
        arg_ids = [arg_tup[1] for arg_tup in self.args]
        # TODO: verify this logic, it's not entirely clear it's right
        if len(arg_ids) > 1:
            for arg in arg_ids:
                soft_deps.add(arg)
        else:
            for arg in arg_ids:
                hard_deps.add(arg)
        return (soft_deps, hard_deps)


class EquivAnnotation(TypedAnnotation):
    """Represents an equivalence group annotation. Equivs define a set of other
    annotations (normally TextBoundAnnotation) to be equivalent.

    Represented in standoff as

    *\tTYPE ID1 ID2 [...]

    Where "*" is the literal asterisk character.
    """

    def __init__(self, type, entities, tail, source_id=None):
        TypedAnnotation.__init__(self, type, tail, source_id=source_id)
        self.entities = entities

    def __in__(self, other):
        return other in self.entities

    def __str__(self):
        return u'*\t%s %s%s' % (
            self.type,
            ' '.join([str(e) for e in self.entities]),
            self.tail
        )

    def get_deps(self):
        soft_deps, hard_deps = TypedAnnotation.get_deps(self)
        if len(self.entities) > 2:
            for ent in self.entities:
                soft_deps.add(ent)
        else:
            for ent in self.entities:
                hard_deps.add(ent)
        return (soft_deps, hard_deps)

    def reference_id(self):
        if self.entities:
            return ['equiv', self.type, self.entities[0]]
        else:
            return ['equiv', self.type, self.entities]

    def reference_text(self):
        return '(' + ','.join([str(e) for e in self.entities]) + ')'


class AttributeAnnotation(IdedAnnotation):
    def __init__(self, target, id, type, tail, value, source_id=None):
        IdedAnnotation.__init__(self, id, type, tail, source_id=source_id)
        self.target = target
        self.value = value

    def __str__(self):
        return u'%s\t%s %s%s%s' % (
            self.id,
            self.type,
            self.target,
            # We hack in old modifiers with this trick using bools
            ' ' + str(self.value) if self.value != True else '',
            self.tail,
        )

    def get_deps(self):
        soft_deps, hard_deps = IdedAnnotation.get_deps(self)
        hard_deps.add(self.target)
        return (soft_deps, hard_deps)

    def reference_id(self):
        # TODO: can't currently ID modifier in isolation; return
        # reference to modified instead
        return [self.target]


class NormalizationAnnotation(IdedAnnotation):
    def __init__(self, _id, _type, target, refdb, refid, tail, source_id=None):
        IdedAnnotation.__init__(self, _id, _type, tail, source_id=source_id)
        self.target = target
        self.refdb = refdb
        self.refid = refid
        # "human-readable" text of referenced ID (optional)
        self.reftext = tail.lstrip('\t').rstrip('\n')

    def __str__(self):
        return u'%s\t%s %s %s:%s%s' % (
            self.id,
            self.type,
            self.target,
            self.refdb,
            self.refid,
            self.tail,
        )

    def get_deps(self):
        soft_deps, hard_deps = IdedAnnotation.get_deps(self)
        hard_deps.add(self.target)
        return (soft_deps, hard_deps)

    def reference_id(self):
        # TODO: can't currently ID normalization in isolation; return
        # reference to target instead
        return [self.target]


class OnelineCommentAnnotation(IdedAnnotation):
    def __init__(self, target, id, type, tail, source_id=None):
        IdedAnnotation.__init__(self, id, type, tail, source_id=source_id)
        self.target = target

    def __str__(self):
        return u'%s\t%s %s%s' % (
            self.id,
            self.type,
            self.target,
            self.tail
        )

    def get_text(self):
        # TODO: will this always hold? Wouldn't it be better to parse
        # further rather than just assuming the whole tail is the text?
        return self.tail.strip()

    def get_deps(self):
        soft_deps, hard_deps = IdedAnnotation.get_deps(self)
        hard_deps.add(self.target)
        return (soft_deps, hard_deps)


class TextBoundAnnotation(IdedAnnotation):
    """Represents a text-bound annotation. Text-bound annotations identify a
    specific span of text and assign it a type.  This base class does not
    assume ability to access text; use TextBoundAnnotationWithText for that.

    Represented in standoff as

    ID\tTYPE START END

    Where START and END are positive integer offsets identifying the
    span of the annotation in text. Discontinuous annotations can be
    represented as

    ID\tTYPE START1 END1;START2 END2;...

    with multiple START END pairs separated by semicolons.
    """

    def __init__(self, spans, id, type, tail, source_id=None):
        # Note: if present, the text goes into tail
        IdedAnnotation.__init__(self, id, type, tail, source_id=source_id)
        self.spans = spans

    # TODO: temp hack while building support for discontinuous
    # annotations; remove once done
    def get_start(self):
        Messager.warning('TextBoundAnnotation.start access')
        return self.spans[0][0]

    def get_end(self):
        Messager.warning('TextBoundAnnotation.end access')
        return self.spans[-1][1]
    start = property(get_start)
    end = property(get_end)
    # end hack

    def first_start(self):
        """Return the first (min) start offset in the annotation spans."""
        return min([start for start, end in self.spans])

    def last_end(self):
        """Return the last (max) end offset in the annotation spans."""
        return max([end for start, end in self.spans])

    def get_text(self):
        # If you're seeing this exception, you probably need a
        # TextBoundAnnotationWithText. The underlying issue may be
        # that you're creating an Annotations object instead of
        # TextAnnotations.
        raise NotImplementedError

    def same_span(self, other):
        """Determine if a given other TextBoundAnnotation has the same span as
        this one.

        Returns True if each (start, end) span of the other annotation
        is equivalent with at least one span of this annotation, False
        otherwise.
        """
        return set(self.spans) == set(other.spans)

    def contains(self, other):
        """Determine if a given other TextBoundAnnotation is contained in this
        one.

        Returns True if each (start, end) span of the other annotation
        is inside (or equivalent with) at least one span of this
        annotation, False otherwise.
        """
        for o_start, o_end in other.spans:
            contained = False
            for s_start, s_end in self.spans:
                if o_start >= s_start and o_end <= s_end:
                    contained = True
                    break
            if not contained:
                return False
        return True

    def __str__(self):
        return u'%s\t%s %s%s' % (
            self.id,
            self.type,
            ';'.join(['%d %d' % (start, end) for start, end in self.spans]),
            self.tail
        )


class TextBoundAnnotationWithText(TextBoundAnnotation):
    """Represents a text-bound annotation. Text-bound annotations identify a
    specific span of text and assign it a type.  This class assume that the
    referenced text is included in the annotation.

    Represented in standoff as

    ID\tTYPE START END\tTEXT

    Where START and END are positive integer offsets identifying the
    span of the annotation in text and TEXT is the corresponding text.
    Discontinuous annotations can be represented as

    ID\tTYPE START1 END1;START2 END2;...

    with multiple START END pairs separated by semicolons.
    """

    def __init__(self, spans, id, type, text, text_tail="", source_id=None):
        IdedAnnotation.__init__(
            self,
            id,
            type,
            '\t' +
            text +
            text_tail,
            source_id=source_id)
        self.spans = spans
        self.text = text
        self.text_tail = text_tail

    # TODO: temp hack while building support for discontinuous
    # annotations; remove once done
    def get_start(self):
        Messager.warning('TextBoundAnnotationWithText.start access')
        return self.spans[0][0]

    def get_end(self):
        Messager.warning('TextBoundAnnotationWithText.end access')
        return self.spans[-1][1]
    start = property(get_start)
    end = property(get_end)
    # end hack

    def get_text(self):
        return self.text

    def __str__(self):
        #log_info('TextBoundAnnotationWithText: __str__: "%s"' % self.text)
        return u'%s\t%s %s\t%s%s' % (
            self.id,
            self.type,
            ';'.join(['%d %d' % (start, end) for start, end in self.spans]),
            self.text,
            self.text_tail
        )


class BinaryRelationAnnotation(IdedAnnotation):
    """Represents a typed binary relation annotation. Relations are assumed not
    to be symmetric (i.e are "directed"); for equivalence relations,
    EquivAnnotation is likely to be more appropriate. Unlike events, relations
    are not associated with text expressions (triggers) stating them.

    Represented in standoff as

    ID\tTYPE ARG1:ID1 ARG2:ID2

    Where ARG1 and ARG2 are arbitrary (but not identical) labels.
    """

    def __init__(
            self,
            id,
            type,
            arg1l,
            arg1,
            arg2l,
            arg2,
            tail,
            source_id=None):
        IdedAnnotation.__init__(self, id, type, tail, source_id=source_id)
        self.arg1l = arg1l
        self.arg1 = arg1
        self.arg2l = arg2l
        self.arg2 = arg2

    def __str__(self):
        return u'%s\t%s %s:%s %s:%s%s' % (
            self.id,
            self.type,
            self.arg1l,
            self.arg1,
            self.arg2l,
            self.arg2,
            self.tail
        )

    def get_deps(self):
        soft_deps, hard_deps = IdedAnnotation.get_deps(self)
        hard_deps.add(self.arg1)
        hard_deps.add(self.arg2)
        return soft_deps, hard_deps


if __name__ == '__main__':
    from sys import stderr, argv
    for ann_path_i, ann_path in enumerate(argv[1:]):
        print(("%s.) '%s' " % (ann_path_i, ann_path, )
                          ).ljust(80, '#'), file=stderr)
        try:
            with Annotations(ann_path) as anns:
                for ann in anns:
                    print(type(ann))
                    print(str(ann).rstrip('\n'), file=stderr)
        except ImportError:
            # Will try to load the config, probably not available
            pass



#!/usr/bin/env python
# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; coding: utf-8; -*-
# vim:set ft=python ts=4 sw=4 sts=4 autoindent:


"""Per-project configuration functionality for Brat Rapid Annotation Tool
(brat)

Author:     Pontus Stenetorp    <pontus is s u-tokyo ac jp>
Author:     Sampo Pyysalo       <smp is s u-tokyo ac jp>
Author:     Illes Solt          <solt tmit bme hu>
Version:    2011-08-15
"""

import re
import sys
import urllib.parse  # TODO reduce scope
import urllib.robotparser  # TODO reduce scope

from bratsubset.annotation import open_textfile
from bratsubset.message import Messager

ENTITY_CATEGORY, EVENT_CATEGORY, RELATION_CATEGORY, UNKNOWN_CATEGORY = range(
    4)


class InvalidProjectConfigException(Exception):
    pass


# names of files in which various configs are found
__access_control_filename = 'acl.conf'
__annotation_config_filename = 'annotation.conf'
__visual_config_filename = 'visual.conf'
__tools_config_filename = 'tools.conf'
__kb_shortcut_filename = 'kb_shortcuts.conf'

# annotation config section name constants
ENTITY_SECTION = "entities"
RELATION_SECTION = "relations"
EVENT_SECTION = "events"
ATTRIBUTE_SECTION = "attributes"

# aliases for config section names
SECTION_ALIAS = {
    "spans": ENTITY_SECTION,
}

__expected_annotation_sections = (
    ENTITY_SECTION,
    RELATION_SECTION,
    EVENT_SECTION,
    ATTRIBUTE_SECTION)
__optional_annotation_sections = []

# visual config section name constants
OPTIONS_SECTION = "options"
LABEL_SECTION = "labels"
DRAWING_SECTION = "drawing"

__expected_visual_sections = (OPTIONS_SECTION, LABEL_SECTION, DRAWING_SECTION)
__optional_visual_sections = [OPTIONS_SECTION]

# tools config section name constants
SEARCH_SECTION = "search"
ANNOTATORS_SECTION = "annotators"
DISAMBIGUATORS_SECTION = "disambiguators"
NORMALIZATION_SECTION = "normalization"

__expected_tools_sections = (
    OPTIONS_SECTION,
    SEARCH_SECTION,
    ANNOTATORS_SECTION,
    DISAMBIGUATORS_SECTION,
    NORMALIZATION_SECTION)
__optional_tools_sections = (
    OPTIONS_SECTION,
    SEARCH_SECTION,
    ANNOTATORS_SECTION,
    DISAMBIGUATORS_SECTION,
    NORMALIZATION_SECTION)

# special relation types for marking which spans can overlap
# ENTITY_NESTING_TYPE used up to version 1.3, now deprecated
ENTITY_NESTING_TYPE = "ENTITY-NESTING"
# TEXTBOUND_OVERLAP_TYPE used from version 1.3 onward
TEXTBOUND_OVERLAP_TYPE = "<OVERLAP>"
SPECIAL_RELATION_TYPES = set([ENTITY_NESTING_TYPE,
                              TEXTBOUND_OVERLAP_TYPE])
OVERLAP_TYPE_ARG = '<OVL-TYPE>'

# visual config default value names
VISUAL_SPAN_DEFAULT = "SPAN_DEFAULT"
VISUAL_ARC_DEFAULT = "ARC_DEFAULT"
VISUAL_ATTR_DEFAULT = "ATTRIBUTE_DEFAULT"

# visual config attribute name lists
SPAN_DRAWING_ATTRIBUTES = ['fgColor', 'bgColor', 'borderColor']
ARC_DRAWING_ATTRIBUTES = ['color', 'dashArray', 'arrowHead', 'labelArrow']
ATTR_DRAWING_ATTRIBUTES = [
    'glyphColor',
    'box',
    'dashArray',
    'glyph',
    'position']

# fallback defaults if config files not found
__default_configuration = """
[entities]
Protein

[relations]
Equiv	Arg1:Protein, Arg2:Protein, <REL-TYPE>:symmetric-transitive

[events]
Protein_binding|GO:0005515	Theme+:Protein
Gene_expression|GO:0010467	Theme:Protein

[attributes]
Negation	Arg:<EVENT>
Speculation	Arg:<EVENT>
"""

__default_visual = """
[labels]
Protein | Protein | Pro | P
Protein_binding | Protein binding | Binding | Bind
Gene_expression | Gene expression | Expression | Exp
Theme | Theme | Th

[drawing]
Protein	bgColor:#7fa2ff
SPAN_DEFAULT	fgColor:black, bgColor:lightgreen, borderColor:black
ARC_DEFAULT	color:black
ATTRIBUTE_DEFAULT	glyph:*
"""

__default_tools = """
[search]
google     <URL>:http://www.google.com/search?q=%s
"""

__default_kb_shortcuts = """
P	Protein
"""

__default_access_control = """
User-agent: *
Allow: /
Disallow: /hidden/

User-agent: guest
Disallow: /confidential/
"""

# Reserved strings with special meanings in configuration.
reserved_config_name = [
    "ANY",
    "ENTITY",
    "RELATION",
    "EVENT",
    "NONE",
    "EMPTY",
    "REL-TYPE",
    "URL",
    "URLBASE",
    "GLYPH-POS",
    "DEFAULT",
    "NORM",
    "OVERLAP",
    "OVL-TYPE",
    "INHERIT"]
# TODO: "GLYPH-POS" is no longer used, warn if encountered and
# recommend to use "position" instead.
reserved_config_string = ["<%s>" % n for n in reserved_config_name]

# Magic string to use to represent a separator in a config
SEPARATOR_STR = "SEPARATOR"


def normalize_to_storage_form(t):
    """Given a label, returns a form of the term that can be used for disk
    storage.

    For example, space can be replaced with underscores to allow use
    with space-separated formats.
    """
    if t not in normalize_to_storage_form.__cache:
        # conservative implementation: replace any space with
        # underscore, replace unicode accented characters with
        # non-accented equivalents, remove others, and finally replace
        # all characters not in [a-zA-Z0-9_-] with underscores.

        import re
        import unicodedata

        n = t.replace(" ", "_")
        if isinstance(n, str):
            unicodedata.normalize('NFKD', n).encode('ascii', 'ignore')
        n = re.sub(r'[^a-zA-Z0-9_-]', '_', n)

        normalize_to_storage_form.__cache[t] = n

    return normalize_to_storage_form.__cache[t]


normalize_to_storage_form.__cache = {}


class TypeHierarchyNode:
    """Represents a node in a simple (possibly flat) hierarchy.

    Each node is associated with a set of terms, one of which (the
    storage_form) matches the way in which the type denoted by the
    node is referenced to in data stored on disk and in client-server
    communications. This term is guaranteed to be in "storage form" as
    defined by normalize_to_storage_form().

    Each node may be associated with one or more "arguments", which
    are (multivalued) key:value pairs. These determine various characteristics
    of the node, but their interpretation depends on the hierarchy the
    node occupies: for example, for events the arguments correspond to
    event arguments.
    """

    def __init__(self, terms, args=[]):
        self.terms, self.args = terms, args

        if len(terms) == 0 or len([t for t in terms if t == ""]) != 0:
            Messager.debug("Empty term in configuration", duration=-1)
            raise InvalidProjectConfigException

        # unused if any of the terms marked with "!"
        self.unused = False
        for i in range(len(self.terms)):
            if self.terms[i][0] == "!":
                self.terms[i] = self.terms[i][1:]
                self.unused = True
        self.children = []

        # The first of the listed terms is used as the primary term for
        # storage (excepting for "special" config-only types). Due to
        # format restrictions, this form must not have e.g. space or
        # various special characters.
        if self.terms[0] not in SPECIAL_RELATION_TYPES:
            self.__primary_term = normalize_to_storage_form(self.terms[0])
        else:
            self.__primary_term = self.terms[0]
        # TODO: this might not be the ideal place to put this warning
        if self.__primary_term != self.terms[0]:
            Messager.warning(
                "Note: in configuration, term '%s' is not appropriate for storage (should match '^[a-zA-Z0-9_-]*$'), using '%s' instead. (Revise configuration file to get rid of this message. Terms other than the first are not subject to this restriction.)" %
                (self.terms[0], self.__primary_term), -1)
            self.terms[0] = self.__primary_term

        # TODO: cleaner and more localized parsing
        self.arguments = {}
        self.special_arguments = {}
        self.arg_list = []
        self.arg_min_count = {}
        self.arg_max_count = {}
        self.keys_by_type = {}
        for a in self.args:
            a = a.strip()
            m = re.match(r'^(\S*?):(\S*)$', a)
            if not m:
                Messager.warning(
                    "Project configuration: Failed to parse argument '%s' (args: %s)" %
                    (a, args), 5)
                raise InvalidProjectConfigException
            key, atypes = m.groups()

            # special case (sorry): if the key is a reserved config
            # string (e.g. "<REL-TYPE>" or "<URL>"), parse differently
            # and store separately
            if key in reserved_config_string:
                if key is self.special_arguments:
                    Messager.warning(
                        "Project configuration: error parsing: %s argument '%s' appears multiple times." %
                        key, 5)
                    raise InvalidProjectConfigException
                # special case in special case: relation type specifications
                # are split by hyphens, nothing else is.
                # (really sorry about this.)
                if key == "<REL-TYPE>":
                    self.special_arguments[key] = atypes.split("-")
                else:
                    self.special_arguments[key] = [atypes]
                # NOTE: skip the rest of processing -- don't add in normal args
                continue

            # Parse "repetition" modifiers. These are regex-like:
            # - Arg      : mandatory argument, exactly one
            # - Arg?     : optional argument, at most one
            # - Arg*     : optional argument, any number
            # - Arg+     : mandatory argument, one or more
            # - Arg{N}   : mandatory, exactly N
            # - Arg{N-M} : mandatory, between N and M

            m = re.match(r'^(\S+?)(\{\S+\}|\?|\*|\+|)$', key)
            if not m:
                Messager.warning(
                    "Project configuration: error parsing argument '%s'." %
                    key, 5)
                raise InvalidProjectConfigException
            key, rep = m.groups()

            if rep == '':
                # exactly one
                minimum_count = 1
                maximum_count = 1
            elif rep == '?':
                # zero or one
                minimum_count = 0
                maximum_count = 1
            elif rep == '*':
                # any number
                minimum_count = 0
                maximum_count = sys.maxsize
            elif rep == '+':
                # one or more
                minimum_count = 1
                maximum_count = sys.maxsize
            else:
                # exact number or range constraint
                assert '{' in rep and '}' in rep, "INTERNAL ERROR"
                m = re.match(r'\{(\d+)(?:-(\d+))?\}$', rep)
                if not m:
                    Messager.warning(
                        "Project configuration: error parsing range '%s' in argument '%s' (syntax is '{MIN-MAX}')." %
                        (rep, key + rep), 5)
                    raise InvalidProjectConfigException
                n1, n2 = m.groups()
                n1 = int(n1)
                if n2 is None:
                    # exact number
                    if n1 == 0:
                        Messager.warning(
                            "Project configuration: cannot have exactly 0 repetitions of argument '%s'." %
                            (key + rep), 5)
                        raise InvalidProjectConfigException
                    minimum_count = n1
                    maximum_count = n1
                else:
                    # range
                    n2 = int(n2)
                    if n1 > n2:
                        Messager.warning(
                            "Project configuration: invalid range %d-%d for argument '%s'." %
                            (n1, n2, key + rep), 5)
                        raise InvalidProjectConfigException
                    minimum_count = n1
                    maximum_count = n2

            # format / config sanity: an argument whose label ends
            # with a digit label cannot be repeated, as this would
            # introduce ambiguity into parsing. (For example, the
            # second "Theme" is "Theme2", and the second "Arg1" would
            # be "Arg12".)
            if maximum_count > 1 and key[-1].isdigit():
                Messager.warning(
                    "Project configuration: error parsing: arguments ending with a digit cannot be repeated: '%s'" %
                    (key + rep), 5)
                raise InvalidProjectConfigException

            if key in self.arguments:
                Messager.warning(
                    "Project configuration: error parsing: %s argument '%s' appears multiple times." %
                    key, 5)
                raise InvalidProjectConfigException

            assert (key not in self.arg_min_count and
                    key not in self.arg_max_count), "INTERNAL ERROR"
            self.arg_min_count[key] = minimum_count
            self.arg_max_count[key] = maximum_count

            self.arg_list.append(key)

            for atype in atypes.split("|"):
                if atype.strip() == "":
                    Messager.warning(
                        "Project configuration: error parsing: empty type for argument '%s'." %
                        a, 5)
                    raise InvalidProjectConfigException

                # Check disabled; need to support arbitrary UTF values
                # for visual.conf. TODO: add this check for other configs.
                # TODO: consider checking for similar for appropriate confs.
                #                 if atype not in reserved_config_string and normalize_to_storage_form(atype) != atype:
                #                     Messager.warning("Project configuration: '%s' is not a valid argument (should match '^[a-zA-Z0-9_-]*$')" % atype, 5)
                #                     raise InvalidProjectConfigException

                if key not in self.arguments:
                    self.arguments[key] = []
                self.arguments[key].append(atype)

                if atype not in self.keys_by_type:
                    self.keys_by_type[atype] = []
                self.keys_by_type[atype].append(key)

    def argument_minimum_count(self, arg):
        """Returns the minimum number of times the given argument is required
        to appear for this type."""
        return self.arg_min_count.get(arg, 0)

    def argument_maximum_count(self, arg):
        """Returns the maximum number of times the given argument is allowed to
        appear for this type."""
        return self.arg_max_count.get(arg, 0)

    def mandatory_arguments(self):
        """Returns the arguments that must appear at least once for this
        type."""
        return [a for a in self.arg_list if self.arg_min_count[a] > 0]

    def multiple_allowed_arguments(self):
        """Returns the arguments that may appear multiple times for this
        type."""
        return [a for a in self.arg_list if self.arg_max_count[a] > 1]

    def storage_form(self):
        """Returns the form of the term used for storage serverside."""
        return self.__primary_term

    def normalizations(self):
        """Returns the normalizations applicable to this node, if any."""
        return self.special_arguments.get('<NORM>', [])


def __require_tab_separator(section):
    """Given a section name, returns True iff in that section of the project
    config only tab separators should be permitted.

    This exception initially introduced to allow slighlty different
    syntax for the [labels] section than others.
    """
    return section == "labels"


def __read_term_hierarchy(input, section=None):
    root_nodes = []
    last_node_at_depth = {}
    last_args_at_depth = {}

    macros = {}
    for l in input:
        # skip empties and lines starting with '#'
        if l.strip() == '' or re.match(r'^\s*#', l):
            continue

        # interpret lines of only hyphens as separators
        # for display
        if re.match(r'^\s*-+\s*$', l):
            # TODO: proper placeholder and placing
            root_nodes.append(SEPARATOR_STR)
            continue

        # interpret lines of the format <STR1>=STR2 as "macro"
        # definitions, defining <STR1> as a placeholder that should be
        # replaced with STR2 whevever it occurs.
        m = re.match(r'^<([a-zA-Z_-]+)>=\s*(.*?)\s*$', l)
        if m:
            name, value = m.groups()
            if name in reserved_config_name:
                Messager.error(
                    "Cannot redefine <%s> in configuration, it is a reserved name." %
                    name)
                # TODO: proper exception
                assert False
            else:
                macros["<%s>" % name] = value
            continue

        # macro expansion
        for n in macros:
            l = l.replace(n, macros[n])

        # check for undefined macros
        for m in re.finditer(r'(<.*?>)', l):
            s = m.group(1)
            assert s in reserved_config_string, "Error: undefined macro %s in configuration. (Note that macros are section-specific.)" % s

        # choose strict tab-only separator or looser any-space
        # separator matching depending on section
        if __require_tab_separator(section):
            m = re.match(r'^(\s*)([^\t]+)(?:\t(.*))?$', l)
        else:
            m = re.match(r'^(\s*)(\S+)(?:\s+(.*))?$', l)
        assert m, "Error parsing line: '%s'" % l
        indent, terms, args = m.groups()
        terms = [t.strip() for t in terms.split("|") if t.strip() != ""]
        if args is None or args.strip() == "":
            args = []
        else:
            args = [a.strip() for a in args.split(",") if a.strip() != ""]

        # older configs allowed space in term strings, splitting those
        # from arguments by space. Trying to parse one of these in the
        # new way will result in a crash from space in arguments.
        # The following is a workaround for the transition.
        if len([x for x in args if re.search('\s', x)]) and '\t' in l:
            # re-parse in the old way (dups from above)
            m = re.match(r'^(\s*)([^\t]+)(?:\t(.*))?$', l)
            assert m, "Error parsing line: '%s'" % l
            indent, terms, args = m.groups()
            terms = [t.strip() for t in terms.split("|") if t.strip() != ""]
            if args is None or args.strip() == "":
                args = []
            else:
                args = [a.strip() for a in args.split(",") if a.strip() != ""]
            # issue a warning
            Messager.warning(
                "Space in term name(s) (%s) on line \"%s\" in config. This feature is deprecated and support will be removed in future versions. Please revise your configuration." %
                (",".join(
                    [
                        '"%s"' %
                        x for x in terms if " " in x]),
                 l),
                20)

        # depth in the ontology corresponds to the number of
        # spaces in the initial indent.
        depth = len(indent)

        # expand <INHERIT> into parent arguments
        expanded_args = []
        for a in args:
            if a != '<INHERIT>':
                expanded_args.append(a)
            else:
                assert depth - 1 in last_args_at_depth, \
                    "Error no parent for '%s'" % l
                expanded_args.extend(last_args_at_depth[depth - 1])
        # TODO: remove, debugging
        #         if expanded_args != args:
        #             Messager.info('expand: %s --> %s' % (str(args), str(expanded_args)))
        args = expanded_args

        n = TypeHierarchyNode(terms, args)
        if depth == 0:
            # root level, no children assignments
            root_nodes.append(n)
        else:
            # assign as child of last node at the depth of the parent
            assert depth - 1 in last_node_at_depth, \
                "Error: no parent for '%s'" % l
            last_node_at_depth[depth - 1].children.append(n)
        last_node_at_depth[depth] = n
        last_args_at_depth[depth] = args

    return root_nodes


def __read_or_default(filename, default):
    try:
        f = open_textfile(filename, 'r')
        r = f.read()
        f.close()
        return r
    except BaseException:
        # TODO: specific exception handling and reporting
        return default


def __parse_kb_shortcuts(shortcutstr, default, source):
    try:
        shortcuts = {}
        for l in shortcutstr.split("\n"):
            l = l.strip()
            if l == "" or l[:1] == "#":
                continue
            key, type = re.split(r'[ \t]+', l)
            if key in shortcuts:
                Messager.warning(
                    "Project configuration: keyboard shortcut for '%s' defined multiple times. Ignoring all but first ('%s')" %
                    (key, shortcuts[key]))
            else:
                shortcuts[key] = type
    except BaseException:
        # TODO: specific exception handling
        Messager.warning(
            "Project configuration: error parsing keyboard shortcuts from %s. Configuration may be wrong." %
            source, 5)
        shortcuts = default
    return shortcuts


def __parse_access_control(acstr, source):
    try:
        parser = urllib.robotparser.RobotFileParser()
        parser.parse(acstr.split("\n"))
    except BaseException:
        # TODO: specific exception handling
        Messager.warning(
            "Project configuration: error parsing access control rules from %s. Configuration may be wrong." %
            source, 5)
        parser = None
    return parser


def get_config_path(directory):
    return __read_first_in_directory_tree(
        directory, __annotation_config_filename)[1]


def __read_first_in_directory_tree(directory, filename):
    # config will not be available command-line invocations;
    # in these cases search whole tree
    BASE_DIR = "/"
    from os.path import split, join

    source, result = None, None

    # check from the given directory and parents, but not above BASE_DIR
    if directory is not None:
        # TODO: this check may fail; consider "foo//bar/data"
        while BASE_DIR in directory:
            source = join(directory, filename)
            result = __read_or_default(source, None)
            if result is not None:
                break
            directory = split(directory)[0]

    return (result, source)


def __parse_configs(configstr, source, expected_sections, optional_sections):
    # top-level config structure is a set of term hierarchies
    # separated by lines consisting of "[SECTION]" where SECTION is
    # e.g.  "entities", "relations", etc.

    # start by splitting config file lines by section, also storing
    # the label (default name or alias) used for each section.

    section = "general"
    section_lines = {section: []}
    section_labels = {}
    for ln, l in enumerate(configstr.split("\n")):
        m = re.match(r'^\s*\[(.*)\]\s*$', l)
        if m:
            section = m.group(1)

            # map and store section name/alias (e.g. "spans" -> "entities")
            section_name = SECTION_ALIAS.get(section, section)
            section_labels[section_name] = section
            section = section_name

            if section not in expected_sections:
                Messager.warning(
                    "Project configuration: unexpected section [%s] in %s. Ignoring contents." %
                    (section, source), 5)
            if section not in section_lines:
                section_lines[section] = []
        else:
            section_lines[section].append(l)

    # attempt to parse lines in each section as a term hierarchy
    configs = {}
    for s, sl in list(section_lines.items()):
        try:
            configs[s] = __read_term_hierarchy(sl, s)
        except Exception as e:
            Messager.warning(
                "Project configuration: error parsing section [%s] in %s: %s" %
                (s, source, str(e)), 5)
            raise

    # verify that expected sections are present; replace with empty if not.
    for s in expected_sections:
        if s not in configs:
            if s not in optional_sections:
                Messager.warning(
                    "Project configuration: missing section [%s] in %s. Configuration may be wrong." %
                    (s, source), 5)
            configs[s] = []

    return (configs, section_labels)


def get_configs(
        directory,
        filename,
        defaultstr,
        minconf,
        sections,
        optional_sections):
    if (directory, filename) not in get_configs.__cache:
        configstr, source = __read_first_in_directory_tree(directory, filename)

        if configstr is None:
            # didn't get one; try default dir and fall back to the default
            configstr = __read_or_default(filename, defaultstr)
            if configstr == defaultstr:
                Messager.info(
                    "Project configuration: no configuration file (%s) found, using default." %
                    filename, 5)
                source = "[default]"
            else:
                source = filename

        # try to parse what was found, fall back to minimal config
        try:
            configs, section_labels = __parse_configs(
                configstr, source, sections, optional_sections)
        except BaseException:
            Messager.warning(
                "Project configuration: Falling back to minimal default. Configuration is likely wrong.",
                5)
            configs = minconf
            section_labels = dict([(a, a) for a in sections])

        # very, very special case processing: if we have a type
        # "Equiv" defined in a "relations" section that doesn't
        # specify a "<REL-TYPE>", automatically fill "symmetric" and
        # "transitive". This is to support older configurations that
        # rely on the type "Equiv" to identify the relation as an
        # equivalence.
        if 'relations' in configs:
            for r in configs['relations']:
                if r == SEPARATOR_STR:
                    continue
                if (r.storage_form() == "Equiv" and
                        "<REL-TYPE>" not in r.special_arguments):
                    # this was way too much noise; will only add in after
                    # at least most configs are revised.
                    #                     Messager.warning('Note: "Equiv" defined in config without "<REL-TYPE>"; assuming symmetric and transitive. Consider revising config to add "<REL-TYPE>:symmetric-transitive" to definition.')
                    r.special_arguments["<REL-TYPE>"] = ["symmetric",
                                                         "transitive"]

        get_configs.__cache[(directory, filename)] = (configs, section_labels)

    return get_configs.__cache[(directory, filename)]


get_configs.__cache = {}


def __get_access_control(directory, filename, default_rules):
    acstr, source = __read_first_in_directory_tree(directory, filename)

    if acstr is None:
        acstr = default_rules  # TODO read or default isntead of default
        if acstr == default_rules:
            source = "[default rules]"
        else:
            source = filename
    ac_oracle = __parse_access_control(acstr, source)
    return ac_oracle


def __get_kb_shortcuts(directory, filename, default_shortcuts, min_shortcuts):
    shortcutstr, source = __read_first_in_directory_tree(directory, filename)

    if shortcutstr is None:
        shortcutstr = __read_or_default(filename, default_shortcuts)
        if shortcutstr == default_shortcuts:
            source = "[default kb_shortcuts]"
        else:
            source = filename

    kb_shortcuts = __parse_kb_shortcuts(shortcutstr, min_shortcuts, source)
    return kb_shortcuts


# final fallback for configuration; a minimal known-good config
__minimal_configuration = {
    ENTITY_SECTION: [
        TypeHierarchyNode(
            ["Protein"])], RELATION_SECTION: [
        TypeHierarchyNode(
            ["Equiv"], [
                "Arg1:Protein", "Arg2:Protein", "<REL-TYPE>:symmetric-transitive"])], EVENT_SECTION: [
        TypeHierarchyNode(
            ["Event"], ["Theme:Protein"])], ATTRIBUTE_SECTION: [
        TypeHierarchyNode(
            ["Negation"], ["Arg:<EVENT>"])], }


def get_annotation_configs(directory):
    return get_configs(directory,
                       __annotation_config_filename,
                       __default_configuration,
                       __minimal_configuration,
                       __expected_annotation_sections,
                       __optional_annotation_sections)


# final fallback for visual configuration; minimal known-good config
__minimal_visual = {
    LABEL_SECTION: [TypeHierarchyNode(["Protein", "Pro", "P"]),
                    TypeHierarchyNode(["Equiv", "Eq"]),
                    TypeHierarchyNode(["Event", "Ev"])],
    DRAWING_SECTION: [TypeHierarchyNode([VISUAL_SPAN_DEFAULT], ["fgColor:black", "bgColor:white"]),
                      TypeHierarchyNode([VISUAL_ARC_DEFAULT], ["color:black"]),
                      TypeHierarchyNode([VISUAL_ATTR_DEFAULT], ["glyph:*"])],
}


def get_visual_configs(directory):
    return get_configs(directory,
                       __visual_config_filename,
                       __default_visual,
                       __minimal_visual,
                       __expected_visual_sections,
                       __optional_visual_sections)


# final fallback for tools configuration; minimal known-good config
__minimal_tools = {
    OPTIONS_SECTION: [],
    SEARCH_SECTION: [
        TypeHierarchyNode(
            ["google"],
            ["<URL>:http://www.google.com/search?q=%s"])],
    ANNOTATORS_SECTION: [],
    DISAMBIGUATORS_SECTION: [],
    NORMALIZATION_SECTION: [],
}


def get_tools_configs(directory):
    return get_configs(directory,
                       __tools_config_filename,
                       __default_tools,
                       __minimal_tools,
                       __expected_tools_sections,
                       __optional_tools_sections)


def get_entity_type_hierarchy(directory):
    return get_annotation_configs(directory)[0][ENTITY_SECTION]


def get_relation_type_hierarchy(directory):
    return get_annotation_configs(directory)[0][RELATION_SECTION]


def get_event_type_hierarchy(directory):
    return get_annotation_configs(directory)[0][EVENT_SECTION]


def get_attribute_type_hierarchy(directory):
    return get_annotation_configs(directory)[0][ATTRIBUTE_SECTION]


def get_annotation_config_section_labels(directory):
    return get_annotation_configs(directory)[1]


# TODO: too much caching?


def get_labels(directory):
    cache = get_labels.__cache
    if directory not in cache:
        l = {}
        for t in get_visual_configs(directory)[0][LABEL_SECTION]:
            if t.storage_form() in l:
                Messager.warning(
                    "In configuration, labels for '%s' defined more than once. Only using the last set." %
                    t.storage_form(), -1)
            # first is storage for, rest are labels.
            l[t.storage_form()] = t.terms[1:]
        cache[directory] = l
    return cache[directory]


get_labels.__cache = {}


# TODO: too much caching?


def get_drawing_types(directory):
    cache = get_drawing_types.__cache
    if directory not in cache:
        l = set()
        for n in get_drawing_config(directory):
            l.add(n.storage_form())
        cache[directory] = list(l)
    return cache[directory]


get_drawing_types.__cache = {}


def get_option_config(directory):
    return get_tools_configs(directory)[0][OPTIONS_SECTION]


def get_drawing_config(directory):
    return get_visual_configs(directory)[0][DRAWING_SECTION]


def get_visual_option_config(directory):
    return get_visual_configs(directory)[0][OPTIONS_SECTION]


def get_visual_config_section_labels(directory):
    return get_visual_configs(directory)[1]


def get_search_config(directory):
    return get_tools_configs(directory)[0][SEARCH_SECTION]


def get_annotator_config(directory):
    return get_tools_configs(directory)[0][ANNOTATORS_SECTION]


def get_disambiguator_config(directory):
    return get_tools_configs(directory)[0][DISAMBIGUATORS_SECTION]


def get_normalization_config(directory):
    return get_tools_configs(directory)[0][NORMALIZATION_SECTION]


def get_tools_config_section_labels(directory):
    return get_tools_configs(directory)[1]


def get_access_control(directory):
    cache = get_access_control.__cache
    if directory not in cache:
        a = __get_access_control(directory,
                                 __access_control_filename,
                                 __default_access_control)
        cache[directory] = a

    return cache[directory]


get_access_control.__cache = {}


def get_kb_shortcuts(directory):
    cache = get_kb_shortcuts.__cache
    if directory not in cache:
        a = __get_kb_shortcuts(directory,
                               __kb_shortcut_filename,
                               __default_kb_shortcuts,
                               {"P": "Positive_regulation"})
        cache[directory] = a

    return cache[directory]


get_kb_shortcuts.__cache = {}


def __collect_type_list(node, collected):
    if node == SEPARATOR_STR:
        return collected

    collected.append(node)

    for c in node.children:
        __collect_type_list(c, collected)

    return collected


def __type_hierarchy_to_list(hierarchy):
    root_nodes = hierarchy
    types = []
    for n in root_nodes:
        __collect_type_list(n, types)
    return types


# TODO: it's not clear it makes sense for all of these methods to have
# their own caches; this seems a bit like a case of premature
# optimization to me. Consider simplifying.


def get_entity_type_list(directory):
    cache = get_entity_type_list.__cache
    if directory not in cache:
        cache[directory] = __type_hierarchy_to_list(
            get_entity_type_hierarchy(directory))
    return cache[directory]


get_entity_type_list.__cache = {}


def get_event_type_list(directory):
    cache = get_event_type_list.__cache
    if directory not in cache:
        cache[directory] = __type_hierarchy_to_list(
            get_event_type_hierarchy(directory))
    return cache[directory]


get_event_type_list.__cache = {}


def get_relation_type_list(directory):
    cache = get_relation_type_list.__cache
    if directory not in cache:
        cache[directory] = __type_hierarchy_to_list(
            get_relation_type_hierarchy(directory))
    return cache[directory]


get_relation_type_list.__cache = {}


def get_attribute_type_list(directory):
    cache = get_attribute_type_list.__cache
    if directory not in cache:
        cache[directory] = __type_hierarchy_to_list(
            get_attribute_type_hierarchy(directory))
    return cache[directory]


get_attribute_type_list.__cache = {}


def get_search_config_list(directory):
    cache = get_search_config_list.__cache
    if directory not in cache:
        cache[directory] = __type_hierarchy_to_list(
            get_search_config(directory))
    return cache[directory]


get_search_config_list.__cache = {}


def get_annotator_config_list(directory):
    cache = get_annotator_config_list.__cache
    if directory not in cache:
        cache[directory] = __type_hierarchy_to_list(
            get_annotator_config(directory))
    return cache[directory]


get_annotator_config_list.__cache = {}


def get_disambiguator_config_list(directory):
    cache = get_disambiguator_config_list.__cache
    if directory not in cache:
        cache[directory] = __type_hierarchy_to_list(
            get_disambiguator_config(directory))
    return cache[directory]


get_disambiguator_config_list.__cache = {}


def get_normalization_config_list(directory):
    cache = get_normalization_config_list.__cache
    if directory not in cache:
        cache[directory] = __type_hierarchy_to_list(
            get_normalization_config(directory))
    return cache[directory]


get_normalization_config_list.__cache = {}


def get_node_by_storage_form(directory, term):
    cache = get_node_by_storage_form.__cache
    if directory not in cache:
        d = {}
        for e in get_entity_type_list(
                directory) + get_event_type_list(directory):
            t = e.storage_form()
            if t in d:
                Messager.warning(
                    "Project configuration: term %s appears multiple times, only using last. Configuration may be wrong." %
                    t, 5)
            d[t] = e
        cache[directory] = d

    return cache[directory].get(term, None)


get_node_by_storage_form.__cache = {}


def _get_option_by_storage_form(directory, term, config, cache):
    if directory not in cache:
        d = {}
        for n in config:
            t = n.storage_form()
            if t in d:
                Messager.warning(
                    "Project configuration: %s appears multiple times, only using last. Configuration may be wrong." %
                    t, 5)
            d[t] = {}
            for a in n.arguments:
                if len(n.arguments[a]) != 1:
                    Messager.warning(
                        "Project configuration: %s key %s has multiple values, only using first. Configuration may be wrong." %
                        (t, a), 5)
                d[t][a] = n.arguments[a][0]

        cache[directory] = d

    return cache[directory].get(term, None)


def get_option_config_by_storage_form(directory, term):
    cache = get_option_config_by_storage_form.__cache
    config = get_option_config(directory)
    return _get_option_by_storage_form(directory, term, config, cache)


get_option_config_by_storage_form.__cache = {}


def get_visual_option_config_by_storage_form(directory, term):
    cache = get_visual_option_config_by_storage_form.__cache
    config = get_visual_option_config(directory)
    return _get_option_by_storage_form(directory, term, config, cache)


get_visual_option_config_by_storage_form.__cache = {}


# access for settings for specific options in tools.conf
# TODO: avoid fixed string values here, define vars earlier


def options_get_validation(directory):
    v = get_option_config_by_storage_form(directory, 'Validation')
    return 'none' if v is None else v.get('validate', 'none')


def options_get_tokenization(directory):
    v = get_option_config_by_storage_form(directory, 'Tokens')
    return 'whitespace' if v is None else v.get('tokenizer', 'whitespace')


def options_get_ssplitter(directory):
    v = get_option_config_by_storage_form(directory, 'Sentences')
    return 'regex' if v is None else v.get('splitter', 'regex')


def options_get_annlogfile(directory):
    v = get_option_config_by_storage_form(directory, 'Annotation-log')
    return '<NONE>' if v is None else v.get('logfile', '<NONE>')


# access for settings for specific options in visual.conf


def visual_options_get_arc_bundle(directory):
    v = get_visual_option_config_by_storage_form(directory, 'Arcs')
    return 'none' if v is None else v.get('bundle', 'none')


def visual_options_get_text_direction(directory):
    v = get_visual_option_config_by_storage_form(directory, 'Text')
    return 'ltr' if v is None else v.get('direction', 'ltr')


def get_drawing_config_by_storage_form(directory, term):
    cache = get_drawing_config_by_storage_form.__cache
    if directory not in cache:
        d = {}
        for n in get_drawing_config(directory):
            t = n.storage_form()
            if t in d:
                Messager.warning(
                    "Project configuration: term %s appears multiple times, only using last. Configuration may be wrong." %
                    t, 5)
            d[t] = {}
            for a in n.arguments:
                # attribute drawing can be specified with multiple
                # values (multi-valued attributes), other parts of
                # drawing config should have single values only.
                if len(n.arguments[a]) != 1:
                    if a in ATTR_DRAWING_ATTRIBUTES:
                        # use multi-valued directly
                        d[t][a] = n.arguments[a]
                    else:
                        # warn and pass
                        Messager.warning(
                            "Project configuration: expected single value for %s argument %s, got '%s'. Configuration may be wrong." %
                            (t, a, "|".join(
                                n.arguments[a])))
                else:
                    d[t][a] = n.arguments[a][0]

        # TODO: hack to get around inability to have commas in values;
        # fix original issue instead
        for t in d:
            for k in d[t]:
                # sorry about this
                if not isinstance(d[t][k], list):
                    d[t][k] = d[t][k].replace("-", ",")
                else:
                    for i in range(len(d[t][k])):
                        d[t][k][i] = d[t][k][i].replace("-", ",")

        default_keys = [VISUAL_SPAN_DEFAULT,
                        VISUAL_ARC_DEFAULT,
                        VISUAL_ATTR_DEFAULT]
        for default_dict in [d.get(dk, {}) for dk in default_keys]:
            for k in default_dict:
                for t in d:
                    d[t][k] = d[t].get(k, default_dict[k])

        # Kind of a special case: recognize <NONE> as "deleting" an
        # attribute (prevents default propagation) and <EMPTY> as
        # specifying that a value should be the empty string
        # (can't be written as such directly).
        for t in d:
            todelete = [k for k in d[t] if d[t][k] == '<NONE>']
            for k in todelete:
                del d[t][k]

            for k in d[t]:
                if d[t][k] == '<EMPTY>':
                    d[t][k] = ''

        cache[directory] = d

    return cache[directory].get(term, None)


get_drawing_config_by_storage_form.__cache = {}


def __directory_relations_by_arg_num(
        directory, num, atype, include_special=False):
    assert num >= 0 and num < 2, "INTERNAL ERROR"

    rels = []

    entity_types = set([t.storage_form()
                        for t in get_entity_type_list(directory)])
    event_types = set([t.storage_form()
                       for t in get_event_type_list(directory)])

    for r in get_relation_type_list(directory):
        # "Special" nesting relations ignored unless specifically
        # requested
        if r.storage_form() in SPECIAL_RELATION_TYPES and not include_special:
            continue

        if len(r.arg_list) != 2:
            # Don't complain about argument constraints for unused relations
            if not r.unused:
                Messager.warning(
                    "Relation type %s has %d arguments in configuration (%s; expected 2). Please fix configuration." %
                    (r.storage_form(), len(
                        r.arg_list), ",".join(
                        r.arg_list)))
        else:
            types = r.arguments[r.arg_list[num]]
            for type_ in types:
                # TODO: there has to be a better way
                if (type_ == atype or
                        type_ == "<ANY>" or
                        atype == "<ANY>" or
                        (type_ in entity_types and atype == "<ENTITY>") or
                        (type_ in event_types and atype == "<EVENT>") or
                        (atype in entity_types and type_ == "<ENTITY>") or
                        (atype in event_types and type_ == "<EVENT>")):
                    rels.append(r)
                    # TODO: why not break here?

    return rels


def get_relations_by_arg1(directory, atype, include_special=False):
    cache = get_relations_by_arg1.__cache
    cache[directory] = cache.get(directory, {})
    if (atype, include_special) not in cache[directory]:
        cache[directory][(atype, include_special)] = __directory_relations_by_arg_num(
            directory, 0, atype, include_special)
    return cache[directory][(atype, include_special)]


get_relations_by_arg1.__cache = {}


def get_relations_by_arg2(directory, atype, include_special=False):
    cache = get_relations_by_arg2.__cache
    cache[directory] = cache.get(directory, {})
    if (atype, include_special) not in cache[directory]:
        cache[directory][(atype, include_special)] = __directory_relations_by_arg_num(
            directory, 1, atype, include_special)
    return cache[directory][(atype, include_special)]


get_relations_by_arg2.__cache = {}


def get_relations_by_storage_form(directory, rtype, include_special=False):
    cache = get_relations_by_storage_form.__cache
    cache[directory] = cache.get(directory, {})
    if include_special not in cache[directory]:
        cache[directory][include_special] = {}
        for r in get_relation_type_list(directory):
            if (r.storage_form() in SPECIAL_RELATION_TYPES and
                    not include_special):
                continue
            if r.unused:
                continue
            if r.storage_form() not in cache[directory][include_special]:
                cache[directory][include_special][r.storage_form()] = []
            cache[directory][include_special][r.storage_form()].append(r)
    return cache[directory][include_special].get(rtype, [])


get_relations_by_storage_form.__cache = {}


def get_labels_by_storage_form(directory, term):
    cache = get_labels_by_storage_form.__cache
    if directory not in cache:
        cache[directory] = {}
        for l, labels in list(get_labels(directory).items()):
            # recognize <EMPTY> as specifying that a label should
            # be the empty string
            labels = [lab if lab != '<EMPTY>' else ' ' for lab in labels]
            cache[directory][l] = labels
    return cache[directory].get(term, None)


get_labels_by_storage_form.__cache = {}

# fallback for missing or partial config: these are highly likely to
# be entity (as opposed to an event or relation) types.
# TODO: remove this workaround once the configs stabilize.
very_likely_physical_entity_types = [
    'Protein',
    'Entity',
    'Organism',
    'Chemical',
    'Two-component-system',
    'Regulon-operon',
    # for more PTM annotation
    'Protein_family_or_group',
    'DNA_domain_or_region',
    'Protein_domain_or_region',
    'Amino_acid_monomer',
    'Carbohydrate',
    # for AZ corpus
    'Cell_type',
    'Drug_or_compound',
    'Gene_or_gene_product',
    'Tissue',
    # 'Not_sure',
    # 'Other',
    'Other_pharmaceutical_agent',
]


# helper; doesn't really belong here
# TODO: shouldn't we have an utils.py or something for stuff like this?


def unique_preserve_order(iterable):
    seen = set()
    uniqued = []
    for i in iterable:
        if i not in seen:
            seen.add(i)
            uniqued.append(i)
    return uniqued


class ProjectConfiguration(object):
    def __init__(self, directory):
        # debugging (note: latter test for windows paths)
        if directory[:1] != "/" and not re.search(r'^[a-zA-Z]:\\', directory):
            Messager.debug(
                "Project config received relative directory ('%s'), configuration may not be found." %
                directory, duration=-1)
        self.directory = directory

    def mandatory_arguments(self, atype):
        """Returns the mandatory argument types that must be present for an
        annotation of the given type."""
        node = get_node_by_storage_form(self.directory, atype)
        if node is None:
            Messager.warning(
                "Project configuration: unknown event type %s. Configuration may be wrong." %
                atype)
            return []
        return node.mandatory_arguments()

    def multiple_allowed_arguments(self, atype):
        """Returns the argument types that are allowed to be filled more than
        once for an annotation of the given type."""
        node = get_node_by_storage_form(self.directory, atype)
        if node is None:
            Messager.warning(
                "Project configuration: unknown event type %s. Configuration may be wrong." %
                atype)
            return []
        return node.multiple_allowed_arguments()

    def argument_maximum_count(self, atype, arg):
        """Returns the maximum number of times that the given argument is
        allowed to be filled for an annotation of the given type."""
        node = get_node_by_storage_form(self.directory, atype)
        if node is None:
            Messager.warning(
                "Project configuration: unknown event type %s. Configuration may be wrong." %
                atype)
            return 0
        return node.argument_maximum_count(arg)

    def argument_minimum_count(self, atype, arg):
        """Returns the minimum number of times that the given argument is
        allowed to be filled for an annotation of the given type."""
        node = get_node_by_storage_form(self.directory, atype)
        if node is None:
            Messager.warning(
                "Project configuration: unknown event type %s. Configuration may be wrong." %
                atype)
            return 0
        return node.argument_minimum_count(arg)

    def arc_types_from(self, from_ann):
        return self.arc_types_from_to(from_ann)

    def relation_types_from(self, from_ann, include_special=False):
        """Returns the possible relation types that can have an annotation of
        the given type as their arg1."""
        return [r.storage_form() for r in get_relations_by_arg1(
            self.directory, from_ann, include_special)]

    def relation_types_to(self, to_ann, include_special=False):
        """Returns the possible relation types that can have an annotation of
        the given type as their arg2."""
        return [
            r.storage_form() for r in get_relations_by_arg2(
                self.directory,
                to_ann,
                include_special)]

    def relation_types_from_to(self, from_ann, to_ann, include_special=False):
        """Returns the possible relation types that can have the given arg1 and
        arg2."""
        types = []

        t1r = get_relations_by_arg1(self.directory, from_ann, include_special)
        t2r = get_relations_by_arg2(self.directory, to_ann, include_special)

        for r in t1r:
            if r in t2r:
                types.append(r.storage_form())

        return types

    def overlap_types(self, inner, outer):
        """Returns the set of annotation overlap types that have been
        configured for the given pair of annotations."""
        # TODO: this is O(NM) for relation counts N and M and goes
        # past much of the implemented caching. Might become a
        # bottleneck for annotations with large type systems.
        t1r = get_relations_by_arg1(self.directory, inner, True)
        t2r = get_relations_by_arg2(self.directory, outer, True)

        types = []
        for r in (s for s in t1r if s.storage_form()
                                    in SPECIAL_RELATION_TYPES):
            if r in t2r:
                types.append(r)

        # new-style overlap configuration ("<OVERLAP>") takes precedence
        # over old-style configuration ("ENTITY-NESTING").
        ovl_types = set()

        ovl = [r for r in types if r.storage_form() == TEXTBOUND_OVERLAP_TYPE]
        nst = [r for r in types if r.storage_form() == ENTITY_NESTING_TYPE]

        if ovl:
            if nst:
                Messager.warning(
                    'Warning: both ' +
                    TEXTBOUND_OVERLAP_TYPE +
                    ' and ' +
                    ENTITY_NESTING_TYPE +
                    ' defined for ' +
                    '(' +
                    inner +
                    ',' +
                    outer +
                    ') in config. ' +
                    'Ignoring latter.')
            for r in ovl:
                if OVERLAP_TYPE_ARG not in r.special_arguments:
                    Messager.warning('Warning: missing ' + OVERLAP_TYPE_ARG +
                                     ' for ' + TEXTBOUND_OVERLAP_TYPE +
                                     ', ignoring specification.')
                    continue
                for val in r.special_arguments[OVERLAP_TYPE_ARG]:
                    ovl_types |= set(val.split('|'))
        elif nst:
            # translate into new-style configuration
            ovl_types = set(['contain'])
        else:
            ovl_types = set()

        undefined_types = [t for t in ovl_types if
                           t not in ('contain', 'equal', 'cross', '<ANY>')]
        if undefined_types:
            Messager.warning('Undefined ' + OVERLAP_TYPE_ARG + ' value(s) ' +
                             str(undefined_types) + ' for ' +
                             '(' + inner + ',' + outer + ') in config. ')
        return ovl_types

    def span_can_contain(self, inner, outer):
        """Returns True if the configuration allows the span of an annotation
        of type inner to (properly) contain an annotation of type outer, False
        otherwise."""
        ovl_types = self.overlap_types(inner, outer)
        if 'contain' in ovl_types or '<ANY>' in ovl_types:
            return True
        ovl_types = self.overlap_types(outer, inner)
        if '<ANY>' in ovl_types:
            return True
        return False

    def spans_can_be_equal(self, t1, t2):
        """Returns True if the configuration allows the spans of annotations of
        type t1 and t2 to be equal, False otherwise."""
        ovl_types = self.overlap_types(t1, t2)
        if 'equal' in ovl_types or '<ANY>' in ovl_types:
            return True
        ovl_types = self.overlap_types(t2, t1)
        if 'equal' in ovl_types or '<ANY>' in ovl_types:
            return True
        return False

    def spans_can_cross(self, t1, t2):
        """Returns True if the configuration allows the spans of annotations of
        type t1 and t2 to cross, False otherwise."""
        ovl_types = self.overlap_types(t1, t2)
        if 'cross' in ovl_types or '<ANY>' in ovl_types:
            return True
        ovl_types = self.overlap_types(t2, t1)
        if 'cross' in ovl_types or '<ANY>' in ovl_types:
            return True
        return False

    def all_connections(self, include_special=False):
        """Returns a dict of dicts of lists, outer dict keyed by entity/event
        type, inner dicts by role/relation type, and lists containing
        entity/event types, representing all possible connections between
        annotations.

        This function is provided to optimize access to the entire
        annotation configuration for passing it to the client and should
        never be used to check for individual connections. The caller
        must not change the contents of the returned collection.
        """

        # TODO: are these uniques really necessary?
        entity_types = unique_preserve_order(self.get_entity_types())
        event_types = unique_preserve_order(self.get_event_types())
        all_types = unique_preserve_order(entity_types + event_types)

        connections = {}

        # TODO: it might be possible to avoid copies like
        # entity_types[:] and all_types[:] here. Consider the
        # possibility.

        for t1 in all_types:
            assert t1 not in connections, "INTERNAL ERROR"
            connections[t1] = {}

            processed_as_relation = {}

            # relations

            rels = get_relations_by_arg1(self.directory, t1, include_special)

            for r in rels:
                a = r.storage_form()

                conns = connections[t1].get(a, [])

                # magic number "1" is for 2nd argument
                args = r.arguments[r.arg_list[1]]

                if "<ANY>" in args:
                    connections[t1][a] = all_types[:]
                else:
                    for t2 in args:
                        if t2 == "<ENTITY>":
                            conns.extend(entity_types)
                        elif t2 == "<EVENT>":
                            conns.extend(event_types)
                        else:
                            conns.append(t2)
                    connections[t1][a] = unique_preserve_order(conns)

                processed_as_relation[a] = True

            # event arguments

            n1 = get_node_by_storage_form(self.directory, t1)

            for a, args in list(n1.arguments.items()):
                if a in processed_as_relation:
                    Messager.warning(
                        "Project configuration: %s appears both as role and relation. Configuration may be wrong." %
                        a)
                    # won't try to resolve
                    continue

                assert a not in connections[t1], "INTERNAL ERROR"

                # TODO: dedup w/above
                if "<ANY>" in args:
                    connections[t1][a] = all_types[:]
                else:
                    conns = []
                    for t2 in args:
                        if t2 == "<EVENT>":
                            conns.extend(event_types)
                        elif t2 == "<ENTITY>":
                            conns.extend(entity_types)
                        else:
                            conns.append(t2)
                    connections[t1][a] = unique_preserve_order(conns)

        return connections

    def arc_types_from_to(
            self,
            from_ann,
            to_ann="<ANY>",
            include_special=False):
        """Returns the possible arc types that can connect an annotation of
        type from_ann to an annotation of type to_ann.

        If to_ann has the value \"<ANY>\", returns all possible arc
        types.
        """

        from_node = get_node_by_storage_form(self.directory, from_ann)

        if from_node is None:
            Messager.warning(
                "Project configuration: unknown textbound/event type %s. Configuration may be wrong." %
                from_ann)
            return []

        if to_ann == "<ANY>":
            relations_from = get_relations_by_arg1(
                self.directory, from_ann, include_special)
            # TODO: consider using from_node.arg_list instead of .arguments for
            # order
            return unique_preserve_order(
                [role for role in from_node.arguments] + [r.storage_form() for r in relations_from])

        # specific hits
        types = from_node.keys_by_type.get(to_ann, [])

        if "<ANY>" in from_node.keys_by_type:
            types += from_node.keys_by_type["<ANY>"]

        # generic arguments
        if self.is_event_type(to_ann) and '<EVENT>' in from_node.keys_by_type:
            types += from_node.keys_by_type['<EVENT>']
        if self.is_physical_entity_type(
                to_ann) and '<ENTITY>' in from_node.keys_by_type:
            types += from_node.keys_by_type['<ENTITY>']

        # relations
        types.extend(self.relation_types_from_to(from_ann, to_ann))

        return unique_preserve_order(types)

    def attributes_for(self, ann_type):
        """Returs a list of the possible attribute types for an annotation of
        the given type."""
        attrs = []
        for attr in get_attribute_type_list(self.directory):
            if attr == SEPARATOR_STR:
                continue

            if 'Arg' not in attr.arguments:
                Messager.warning(
                    "Project configuration: config error: attribute '%s' lacks 'Arg:' specification." %
                    attr.storage_form())
                continue

            types = attr.arguments['Arg']

            if ((ann_type in types) or ('<ANY>' in types) or
                    (self.is_event_type(ann_type) and '<EVENT>' in types) or
                    (self.is_physical_entity_type(ann_type) and '<ENTITY>' in types)
                    or
                    (self.is_relation_type(ann_type) and '<RELATION>' in types)):
                attrs.append(attr.storage_form())

        return attrs

    def get_labels(self):
        return get_labels(self.directory)

    def get_kb_shortcuts(self):
        return get_kb_shortcuts(self.directory)

    def get_access_control(self):
        return get_access_control(self.directory)

    def get_attribute_types(self):
        return [t.storage_form()
                for t in get_attribute_type_list(self.directory)]

    def get_event_types(self):
        return [t.storage_form() for t in get_event_type_list(self.directory)]

    def get_relation_types(self):
        return [t.storage_form()
                for t in get_relation_type_list(self.directory)]

    def get_equiv_types(self):
        # equivalence relations are those relations that are symmetric
        # and transitive, i.e. that have "symmetric" and "transitive"
        # in their "<REL-TYPE>" special argument values.
        return [t.storage_form() for t in get_relation_type_list(self.directory)
                if "<REL-TYPE>" in t.special_arguments and
                "symmetric" in t.special_arguments["<REL-TYPE>"] and
                "transitive" in t.special_arguments["<REL-TYPE>"]]

    def get_relations_by_type(self, _type):
        return get_relations_by_storage_form(self.directory, _type)

    def get_labels_by_type(self, _type):
        return get_labels_by_storage_form(self.directory, _type)

    def get_drawing_types(self):
        return get_drawing_types(self.directory)

    def get_drawing_config_by_type(self, _type):
        return get_drawing_config_by_storage_form(self.directory, _type)

    def get_search_config(self):
        search_config = []
        for r in get_search_config_list(self.directory):
            if '<URL>' not in r.special_arguments:
                Messager.warning(
                    'Project configuration: config error: missing <URL> specification for %s search.' %
                    r.storage_form())
            else:
                search_config.append(
                    (r.storage_form(), r.special_arguments['<URL>'][0]))
        return search_config

    def _get_tool_config(self, tool_list):
        tool_config = []
        for r in tool_list:
            if '<URL>' not in r.special_arguments:
                Messager.warning(
                    'Project configuration: config error: missing <URL> specification for %s.' %
                    r.storage_form())
                continue
            if 'tool' not in r.arguments:
                Messager.warning(
                    'Project configuration: config error: missing tool name ("tool") for %s.' %
                    r.storage_form())
                continue
            if 'model' not in r.arguments:
                Messager.warning(
                    'Project configuration: config error: missing model name ("model") for %s.' %
                    r.storage_form())
                continue
            tool_config.append((r.storage_form(),
                                r.arguments['tool'][0],
                                r.arguments['model'][0],
                                r.special_arguments['<URL>'][0]))
        return tool_config

    def get_disambiguator_config(self):
        tool_list = get_disambiguator_config_list(self.directory)
        return self._get_tool_config(tool_list)

    def get_annotator_config(self):
        # TODO: "annotator" is a very confusing term for a web service
        # that does automatic annotation in the context of a tool
        # where most annotators are expected to be human. Rethink.
        tool_list = get_annotator_config_list(self.directory)
        return self._get_tool_config(tool_list)

    def get_normalization_config(self):
        norm_list = get_normalization_config_list(self.directory)
        norm_config = []
        for n in norm_list:
            if 'DB' not in n.arguments:
                # optional, server looks in default location if None
                n.arguments['DB'] = [None]
            if '<URL>' not in n.special_arguments:
                Messager.warning(
                    'Project configuration: config error: missing <URL> specification for %s.' %
                    n.storage_form())
                continue
            if '<URLBASE>' not in n.special_arguments:
                # now optional, client skips link generation if None
                n.special_arguments['<URLBASE>'] = [None]
            norm_config.append((n.storage_form(),
                                n.special_arguments['<URL>'][0],
                                n.special_arguments['<URLBASE>'][0],
                                n.arguments['DB'][0]))
        return norm_config

    def get_entity_types(self):
        return [t.storage_form() for t in get_entity_type_list(self.directory)]

    def get_entity_type_hierarchy(self):
        return get_entity_type_hierarchy(self.directory)

    def get_relation_type_hierarchy(self):
        return get_relation_type_hierarchy(self.directory)

    def get_event_type_hierarchy(self):
        return get_event_type_hierarchy(self.directory)

    def get_attribute_type_hierarchy(self):
        return get_attribute_type_hierarchy(self.directory)

    def _get_filtered_attribute_type_hierarchy(self, types):
        from copy import deepcopy
        # TODO: This doesn't property implement recursive traversal
        # and filtering, instead only checking the topmost nodes.
        filtered = []
        for t in self.get_attribute_type_hierarchy():
            if t.storage_form() in types:
                filtered.append(deepcopy(t))
        return filtered

    def attributes_for_types(self, types):
        """Returns list containing the attribute types that are applicable to
        at least one of the given annotation types."""
        # list to preserve order, dict for lookup
        attribute_list = []
        seen = {}
        for t in types:
            for a in self.attributes_for(t):
                if a not in seen:
                    attribute_list.append(a)
                    seen[a] = True
        return attribute_list

    def get_entity_attribute_type_hierarchy(self):
        """Returns the attribute type hierarchy filtered to include only
        attributes that apply to at least one entity."""
        attr_types = self.attributes_for_types(self.get_entity_types())
        return self._get_filtered_attribute_type_hierarchy(attr_types)

    def get_relation_attribute_type_hierarchy(self):
        """Returns the attribute type hierarchy filtered to include only
        attributes that apply to at least one relation."""
        attr_types = self.attributes_for_types(self.get_relation_types())
        return self._get_filtered_attribute_type_hierarchy(attr_types)

    def get_event_attribute_type_hierarchy(self):
        """Returns the attribute type hierarchy filtered to include only
        attributes that apply to at least one event."""
        attr_types = self.attributes_for_types(self.get_event_types())
        return self._get_filtered_attribute_type_hierarchy(attr_types)

    def preferred_display_form(self, t):
        """Given a storage form label, returns the preferred display form as
        defined by the label configuration (labels.conf)"""
        labels = get_labels_by_storage_form(self.directory, t)
        if labels is None or len(labels) < 1:
            return t
        else:
            return labels[0]

    def is_physical_entity_type(self, t):
        if t in self.get_entity_types() or t in self.get_event_types():
            return t in self.get_entity_types()
        # TODO: remove this temporary hack
        if t in very_likely_physical_entity_types:
            return True
        return t in self.get_entity_types()

    def is_event_type(self, t):
        return t in self.get_event_types()

    def is_relation_type(self, t):
        return t in self.get_relation_types()

    def is_equiv_type(self, t):
        return t in self.get_equiv_types()

    def is_configured_type(self, t):
        return (t in self.get_entity_types() or
                t in self.get_event_types() or
                t in self.get_relation_types())

    def type_category(self, t):
        """Returns the category of the given type t.

        The categories can be compared for equivalence but offer no
        other interface.
        """
        if self.is_physical_entity_type(t):
            return ENTITY_CATEGORY
        elif self.is_event_type(t):
            return EVENT_CATEGORY
        elif self.is_relation_type(t):
            return RELATION_CATEGORY
        else:
            # TODO: others
            return UNKNOWN_CATEGORY
